import sys, os
from django.conf import settings
from django.db.backends.creation import BaseDatabaseCreation
from django.db.backends.util import truncate_name
from arches.db.install import install_db

class DatabaseCreation(BaseDatabaseCreation):
    # This dictionary maps Field objects to their associated PostgreSQL column
    # types, as strings. Column-type strings can contain format strings; they'll
    # be interpolated against the values of Field.__dict__ before being output.
    # If a column type is set to None, it won't be included in the output.
    data_types = {
        'AutoField':         'serial',
        'BinaryField':       'bytea',
        'BooleanField':      'boolean',
        'CharField':         'varchar(%(max_length)s)',
        'CommaSeparatedIntegerField': 'varchar(%(max_length)s)',
        'DateField':         'date',
        'DateTimeField':     'timestamp with time zone',
        'DecimalField':      'numeric(%(max_digits)s, %(decimal_places)s)',
        'FileField':         'varchar(%(max_length)s)',
        'FilePathField':     'varchar(%(max_length)s)',
        'FloatField':        'double precision',
        'IntegerField':      'integer',
        'BigIntegerField':   'bigint',
        'IPAddressField':    'inet',
        'GenericIPAddressField': 'inet',
        'NullBooleanField':  'boolean',
        'OneToOneField':     'integer',
        'PositiveIntegerField': 'integer CHECK ("%(column)s" >= 0)',
        'PositiveSmallIntegerField': 'smallint CHECK ("%(column)s" >= 0)',
        'SlugField':         'varchar(%(max_length)s)',
        'SmallIntegerField': 'smallint',
        'TextField':         'text',
        'TimeField':         'time',
    }

    def sql_table_creation_suffix(self):
        assert self.connection.settings_dict['TEST_COLLATION'] is None, "PostgreSQL does not support collation setting at database creation time."
        if self.connection.settings_dict['TEST_CHARSET']:
            return "WITH ENCODING '%s'" % self.connection.settings_dict['TEST_CHARSET']
        return ''

    def sql_indexes_for_field(self, model, f, style):
        output = []
        if f.db_index or f.unique:
            qn = self.connection.ops.quote_name
            db_table = model._meta.db_table
            tablespace = f.db_tablespace or model._meta.db_tablespace
            if tablespace:
                tablespace_sql = self.connection.ops.tablespace_sql(tablespace)
                if tablespace_sql:
                    tablespace_sql = ' ' + tablespace_sql
            else:
                tablespace_sql = ''

            def get_index_sql(index_name, opclass=''):
                return (style.SQL_KEYWORD('CREATE INDEX') + ' ' +
                        style.SQL_TABLE(qn(truncate_name(index_name,self.connection.ops.max_name_length()))) + ' ' +
                        style.SQL_KEYWORD('ON') + ' ' +
                        style.SQL_TABLE(qn(db_table)) + ' ' +
                        "(%s%s)" % (style.SQL_FIELD(qn(f.column)), opclass) +
                        "%s;" % tablespace_sql)

            if not f.unique:
                output = [get_index_sql('%s_%s' % (db_table, f.column))]

            # Fields with database column types of `varchar` and `text` need
            # a second index that specifies their operator class, which is
            # needed when performing correct LIKE queries outside the
            # C locale. See #12234.
            db_type = f.db_type(connection=self.connection)
            if db_type.startswith('varchar'):
                output.append(get_index_sql('%s_%s_like' % (db_table, f.column),
                                            ' varchar_pattern_ops'))
            elif db_type.startswith('text'):
                output.append(get_index_sql('%s_%s_like' % (db_table, f.column),
                                            ' text_pattern_ops'))
        return output

    def create_test_db(self, verbosity=1, autoclobber=False):
        """
        Creates a test database, prompting the user for confirmation if the
        database already exists. Returns the name of the test database created.
        """
        # Don't import django.core.management if it isn't needed.
        from django.core.management import call_command

        test_database_name = self._get_test_db_name()

        if verbosity >= 1:
            test_db_repr = ''
            if verbosity >= 2:
                test_db_repr = " ('%s')" % test_database_name
            print("Creating test database for alias '%s'%s..." % (
                self.connection.alias, test_db_repr))

        self._create_test_db(verbosity, autoclobber)

        self.connection.close()
        settings.DATABASES[self.connection.alias]["NAME"] = test_database_name
        self.connection.settings_dict["NAME"] = test_database_name

        # Report syncdb messages at one level lower than that requested.
        # This ensures we don't get flooded with messages during testing
        # (unless you really ask to be flooded)
        # call_command('syncdb',
        #     verbosity=max(verbosity - 1, 0),
        #     interactive=False,
        #     database=self.connection.alias,
        #     load_initial_data=False)

        # We need to then do a flush to ensure that any data installed by
        # custom SQL has been removed. The only test data should come from
        # test fixtures, or autogenerated from post_syncdb triggers.
        # This has the side effect of loading initial data (which was
        # intentionally skipped in the syncdb).
        # call_command('flush',
        #     verbosity=max(verbosity - 1, 0),
        #     interactive=False,
        #     database=self.connection.alias)

        from django.core.cache import get_cache
        from django.core.cache.backends.db import BaseDatabaseCache
        for cache_alias in settings.CACHES:
            cache = get_cache(cache_alias)
            if isinstance(cache, BaseDatabaseCache):
                call_command('createcachetable', cache._table,
                             database=self.connection.alias)

        # Get a cursor (even though we don't need one yet). This has
        # the side effect of initializing the test database.
        self.connection.cursor()

        return test_database_name

    def _create_test_db(self, verbosity, autoclobber):
        """
        Internal implementation - creates the test db tables.
        """
        suffix = 'test' #self.sql_table_creation_suffix()

        test_database_name = self._get_test_db_name()

        qn = self.connection.ops.quote_name

        try:

            # Create the base arches database and connect to it.
            db_settings = settings.DATABASES['default']
            db_settings['NAME'] = test_database_name
            install_path = os.path.join(settings.ROOT_DIR, 'db', 'install', 'install_db.sql')  
            db_settings['install_path'] = install_path   
            install_db.create_sqlfile(db_settings, install_path)
            
            os.system('psql -h %(HOST)s -p %(PORT)s -U %(USER)s -d postgres -c "DROP DATABASE IF EXISTS %(NAME)s;"' % db_settings)
            os.system('psql -h %(HOST)s -p %(PORT)s -U %(USER)s -d postgres -c "CREATE DATABASE %(NAME)s WITH ENCODING=\'UTF8\' OWNER=%(USER)s CONNECTION LIMIT=-1;"' % db_settings)
            
            os.system('psql -h %(HOST)s -p %(PORT)s -U %(USER)s -d %(NAME)s -f "%(install_path)s"' % db_settings)

        except Exception as e:
            sys.stderr.write(
                "Got an error creating the test database: %s\n" % e)
            if not autoclobber:
                confirm = input(
                    "Type 'yes' if you would like to try deleting the test "
                    "database '%s', or 'no' to cancel: " % test_database_name)
            if autoclobber or confirm == 'yes':
                try:
                    if verbosity >= 1:
                        print("Destroying old test database '%s'..."
                              % self.connection.alias)
                    cursor.execute(
                        "DROP DATABASE %s" % qn(test_database_name))
                    cursor.execute(
                        "CREATE DATABASE %s %s" % (qn(test_database_name),
                                                   suffix))
                except Exception as e:
                    sys.stderr.write(
                        "Got an error recreating the test database: %s\n" % e)
                    sys.exit(2)
            else:
                print("Tests cancelled.")
                sys.exit(1)

        return test_database_name

    def _destroy_test_db(self, test_database_name, verbosity):
        """
        Internal implementation - remove the test db tables.
        """

        db_settings = settings.DATABASES['default']
        db_settings['NAME'] = test_database_name
        os.system('psql -h %(HOST)s -p %(PORT)s -U %(USER)s -d postgres -c "DROP DATABASE IF EXISTS %(NAME)s;"' % db_settings)



from fabric.api import lcd, local
import os, shutil, sys

VIRTUALENV_PATH = os.path.abspath(os.path.join(os.getcwd(), '..', 'virtualenv', 'ENV'))

# SET THE PYTHON.EXE LOCATION
if sys.platform == 'win32':
    PYTHON = os.path.join(VIRTUALENV_PATH, 'Scripts', 'python')
else:
    PYTHON = os.path.join(VIRTUALENV_PATH, 'bin', 'python')

def get_latest():
    """
    Gets the latest code and updates your system

    """

    os.system('hg pull')
    os.system('hg update')

def virtualenv(task='install'):
    with lcd('arches'):
        if task == 'truncate' or task == 'recreate':
            shutil.rmtree(VIRTUALENV_PATH)
        if task == 'install' or task == 'recreate':
            os.system('python setup.py')

def setup_package(package=''):
    """
    Installs Elasticsearch into the package directory and 
    installs the database into postgres as "arches_<package_name>"

    """

    package_opt = '' if package == '' else '--package %s' % package
    os.system('%s manage.py packages --operation setup %s' % (PYTHON, package_opt))

def install_package(package=''):
    """
    Runs the setup.py file found in the package root

    """

    package_opt = '' if package == '' else '--package %s' % package
    os.system('%s manage.py packages --operation install %s' % (PYTHON, package_opt))

def start_es(package=''):
    """
    Start Elasticsearch. Open another window to run more tasks.
    NOTE: this will block all subsequent python calls.  
    You will need to re-run this file to run more tasks

    """

    package_opt = '' if package == '' else '--package %s' % package
    os.system('%s manage.py packages --operation start_elasticsearch %s' % (PYTHON, package_opt))

def setup(package='', virtualenv_task=None):
    """
    Installs Elasticsearch into the package directory and 
    installs the database into postgres as "arches_<package_name>"
    Starts Elasticsearch.
    Runs the setup.py file found in the package root

    """

    if virtualenv_task:
        virtualenv(virtualenv_task)
    setup_package(package=package)
    start_es(package=package)
    install_package(package=package)

def runserver(package='', port=8000):
    """
    Starts the Django web server on a specific port

    """

    os.system('%s manage.py runserver %s -p %s' % (PYTHON, port, package))

def deploy(virtualenv_task='install'):
    """
    Sets up and installs the package from the latest code available

    """

    if virtualenv_task == 'truncate':
        print 'Cannot truncate virtualenv on deploy without recreating -- switching virtualenv_task to "recreate"'
        virtualenv_task = 'recreate'
    get_latest()
    setup(virtualenv_task)

def dev_setup():
    local('python setup.py')
    with lcd('..'):
        local('%s setup.py develop' % (PYTHON))



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

from django.conf import settings
from django.http import HttpResponse
from django.template import RequestContext
from django.shortcuts import render_to_response
from django.views.decorators.csrf import csrf_exempt
from arches.app.models.entity import Entity
from arches.app.utils.betterJSONSerializer import JSONSerializer, JSONDeserializer

@csrf_exempt
def Entities(request, entityid):
    entity = []

    if request.method == 'GET':
        if entityid == '':
            pass
        else:
            entity = Entity().get(entityid)
    else:
        if not request.user.is_authenticated():
            raise Exception('User must be logged in to insert, update, or delete entities')
        if request.method == 'POST':
            if len(request.FILES) > 0:
                jsondata = JSONDeserializer().deserialize(request.POST.get('json'))  
            else:
                jsondata = JSONDeserializer().deserialize(request.body)            
            if not isinstance(jsondata, list):
                jsondata = [jsondata]
            for entityjson in jsondata:
                entity = Entity(entityjson)      
                if entity.entityid != '':
                    entity.delete_index() 
                entity.save(username=request.user.username, uploadedfile=request.FILES.get('file', None))
                entity.index()
        elif request.method == 'DELETE':
            jsondata = JSONDeserializer().deserialize(request.body)
            if not isinstance(jsondata, list):
                jsondata = [jsondata]
            for entityjson in jsondata:
                entity = Entity(entityjson)
                entity.delete_index()                
                entity.delete(delete_root=entity.get_rank()==0)
    return HttpResponse(JSONSerializer().serialize(entity, ensure_ascii=True, indent=4))

@csrf_exempt
def EntityTypes(request, entitytypeid):
    entityschema = []
    if entitytypeid == '':
        return HttpResponse(JSONSerializer().serialize({}, ensure_ascii=True, indent=4))
    else:
        if request.GET.get('f') is None:
            return render_to_response('graph.htm', {}, context_instance=RequestContext(request))
        else:
            entityschema = {entitytypeid: Entity.get_mapping_schema(entitytypeid)}            
            if request.GET.get('f') == 'json':
                return HttpResponse(JSONSerializer().serialize(entityschema, ensure_ascii=True, indent=4), content_type='application/json')
            
            if request.GET.get('f') == 'd3':
                d3Schema = d3Obj()
                d3Schema.name = entitytypeid

                for assestAttr in entityschema[entitytypeid]:
                    d3ObjAssestAttr = d3Obj()
                    d3ObjAssestAttr.name = assestAttr

                    for step in entityschema[entitytypeid][assestAttr]['steps']:
                        d3ObjStep = d3Obj()
                        d3ObjStep.name = step['entitytypedomain'] + ' ' + step['propertyid'] + ' ' + step['entitytyperange'] 
                        d3ObjAssestAttr.children.append(d3ObjStep)
                    d3Schema.children.append(d3ObjAssestAttr)        
                return HttpResponse(JSONSerializer().serialize(d3Schema, ensure_ascii=True, indent=4))

class d3Obj(object):
    def __init__(self):
        self.name = ''
        self.children = []



from PIL import Image, ImageOps
import io
import os
from django.core.files.uploadedfile import SimpleUploadedFile

def generate_thumbnail(file):
	#print file.content_type
	if 'image' in file.content_type:
		im = Image.open(file)
		thumb = ImageOps.fit(im, (128,128), Image.ANTIALIAS)
		f = io.BytesIO()
		thumb.save(f, 'JPEG')
		return SimpleUploadedFile('%s_%s.%s' % ('thumb', os.path.splitext(file.name)[0], 'jpg'), f.getvalue(), content_type='image/jpeg')
	else:
		pass



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

"""This module contains commands for building Arches."""
from optparse import make_option
from django.core.management.base import BaseCommand, CommandError
from django.core import management
from django.conf import settings

class Command(BaseCommand):
    """A general command used in compiling all the required static data files and js file in Arches."""
    
    option_list = BaseCommand.option_list + (
        make_option('-o', '--operation', action='store', dest='operation', default='collectstatic',
            type='choice', choices=['collectstatic'],
            help='Operation Type; collectstatic=Alias for the collectstatic command provided by django'),
    )
    
    def handle(self, *args, **options):
        print 'operation: '+ options['operation']
        if options['operation'] == 'collectstatic':
            if settings.STATIC_ROOT != '':
                management.call_command('collectstatic', interactive=False)



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

from django.conf import settings

def find_candidates(search_string):
    # use search string to retrieve candidate list from data source.
    # return data as follows:
    return [
        {
            'id': 1,
            'text': '100 Main St., San Francisco CA 94109',
            'geometry': {
                "type": "Point",
                "coordinates": [
                    -118.45089,
                    34.08702
                ]
            },
            'score': 99
        },
        {
            'id': 2,
            'text': '101 Main St., San Francisco CA 94109',
            'geometry': {
                "type": "Point",
                "coordinates": [
                    -118.45089,
                    34.08702
                ]
            },
            'score': 99
        },
        {
            'id': 3,
            'text': '102 Main St., San Francisco CA 94109',
            'geometry': {
                "type": "Point",
                "coordinates": [
                    -118.45089,
                    34.08702
                ]
            },
            'score': 99
        },
        {
            'id': 4,
            'text': '103 Main St., San Francisco CA 94109',
            'geometry': {
                "type": "Point",
                "coordinates": [
                    -118.45089,
                    34.08702
                ]
            },
            'score': 99
        }
    ]



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

from django.template import RequestContext
from django.shortcuts import render_to_response, redirect
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.cache import never_cache
from django.contrib.auth import authenticate, login, logout
from django.core.urlresolvers import reverse
from django.conf import settings

def index(request):
    return render_to_response('index.htm', {
            'main_script': 'index',
            'active_page': 'Home',
        },
        context_instance=RequestContext(request))


@never_cache
@csrf_exempt
def auth(request):
    auth_attempt_success = None
    # POST request is taken to mean user is logging in
    if request.method == 'POST':
        username = request.POST['username']
        password = request.POST['password']
        user = authenticate(username=username, password=password)
        if user is not None and user.is_active:
            login(request, user)
            user.password = ''
            auth_attempt_success = True
        else:
            auth_attempt_success = False
    
    next = request.GET.get('next', reverse('home'))
    if auth_attempt_success:
        return redirect(next)
    else:
        if request.GET.get('logout', None) is not None:
            logout(request)

        return render_to_response('login.htm', {
                'main_script': 'login',
                'auth_failed': (auth_attempt_success is not None),
                'next': next
            },
            context_instance=RequestContext(request))


def search(request):
    return render_to_response('search.htm', context_instance=RequestContext(request))



from io import StringIO
from django.http import HttpResponse
from arches.app.utils.betterJSONSerializer import JSONSerializer

class JSONResponse(HttpResponse):

    def __init__(self, content=b'', *args, **kwargs):
    	kwargs['content_type'] = 'text/json'

    	ensure_ascii = kwargs.pop("ensure_ascii", True)
        stream = kwargs.pop("stream", None)
        indent = kwargs.pop("indent", None)
        selected_fields = kwargs.pop("fields", None)
        use_natural_keys = kwargs.pop("use_natural_keys", None)
        geom_format = kwargs.pop("geom_format", None)

        super(HttpResponse, self).__init__(*args, **kwargs)  

        options = {}
        if ensure_ascii != None:
        	options['ensure_ascii'] = ensure_ascii
        if stream != None:
        	options['stream'] = stream
    	if indent != None:
        	options['indent'] = indent
    	if selected_fields != None:
        	options['selected_fields'] = selected_fields
    	if use_natural_keys != None:
        	options['use_natural_keys'] = use_natural_keys
        if geom_format != None:
        	options['geom_format'] = geom_format

        # Content is a bytestring. See the `content` property methods.
        self.content = JSONSerializer().serialize(content, **options)


"""
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import uuid
from django.db import transaction
from django.utils.http import urlencode
from rdflib import Literal, Namespace, RDF, URIRef
from rdflib.namespace import SKOS, DCTERMS
from rdflib.graph import Graph
from time import time
from arches.app.models.concept import Concept
from arches.app.models import models

class SKOSReader(object):
    def __init__(self):
        self.nodes = []
        self.relations = []
            
    def read_file(self, path_to_file, format='xml'):
        """
        parse the skos file and extract all available data

        """       
        rdf_graph = Graph()
        start = time()
        try:
            rdf = rdf_graph.parse(source=path_to_file, format=format)
            print 'time elapsed to parse rdf graph %s s'%(time()-start)
        except:
            raise Exception('Error occurred while parsing the file %s'%path_to_file)
        return rdf
    
    def save_concepts_from_skos(self, graph, concept_scheme_group='00000000-0000-0000-0000-000000000006'):
        """
        given an RDF graph, tries to save the concpets to the system

        """

        baseuuid = uuid.uuid4()
        allowed_languages = models.DLanguages.objects.values_list('pk', flat=True)

        value_types = models.ValueTypes.objects.all()
        skos_value_types = value_types.filter(namespace = 'skos')
        skos_value_types_list = skos_value_types.values_list('valuetype', flat=True)
        dcterms_value_types = value_types.filter(namespace = 'dcterms')

        relation_types = models.DRelationtypes.objects.all()
        skos_relation_types = relation_types.filter(namespace = 'skos')

        
        # if the graph is of the type rdflib.graph.Graph
        if isinstance(graph, Graph):

            # Search for ConceptSchemes first
            for s, v, o in graph.triples((None, RDF.type , SKOS.ConceptScheme)):
                scheme_id = self.generate_uuid_from_subject(baseuuid, s)
                concept = Concept({
                    'id': scheme_id,
                    'legacyoid': str(s),
                    'nodetype': 'ConceptScheme'
                })

                # associate the ConceptScheme with the provided ConceptSchemeGroup 
                self.relations.append({'source': concept_scheme_group, 'type': 'narrower', 'target': scheme_id})
                
                for predicate, object in graph.predicate_objects(subject = s):
                    if str(DCTERMS) in predicate and predicate.replace(DCTERMS, '') in dcterms_value_types.values_list('valuetype', flat=True):
                        if hasattr(object, 'language') and object.language not in allowed_languages: 
                            newlang = models.DLanguages()
                            newlang.pk = object.language
                            newlang.languagename = object.language
                            newlang.isdefault = False
                            newlang.save()
                            allowed_languages = models.DLanguages.objects.values_list('pk', flat=True)

                        try:
                            # first try and get any values associated with the concept
                            value_type = dcterms_value_types.get(valuetype=predicate.replace(DCTERMS, '')) # predicate.replace(SKOS, '') should yield something like 'prefLabel' or 'scopeNote', etc..
                            if predicate == DCTERMS.title:
                                concept.addvalue({'value':object, 'language': object.language, 'type': 'prefLabel', 'category': value_type.category}) 
                                print 'Casting dcterms:title to skos:prefLabel'
                            if predicate == DCTERMS.description:
                                concept.addvalue({'value':object, 'language': object.language, 'type': 'scopeNote', 'category': value_type.category}) 
                                print 'Casting dcterms:description to skos:scopeNote'
                        except:
                            pass

                    if str(SKOS) in predicate:
                        if predicate == SKOS.hasTopConcept:
                            self.relations.append({'source': scheme_id, 'type': 'narrower', 'target': self.generate_uuid_from_subject(baseuuid, object)})

                self.nodes.append(concept)   

            if len(self.nodes) == 0:
                raise Exception('No ConceptScheme found in file.')         

            # Search for Concepts
            for s, v, o in graph.triples((None, RDF.type , SKOS.Concept)):
                concept = Concept({
                    'id': self.generate_uuid_from_subject(baseuuid, s),
                    'legacyoid': str(s),
                    'nodetype': 'Concept'
                })

                # loop through all the elements within a <skos:Concept> element
                for predicate, object in graph.predicate_objects(subject = s):
                    if str(SKOS) in predicate:
                        if hasattr(object, 'language') and object.language not in allowed_languages: 
                            newlang = models.DLanguages()
                            newlang.pk = object.language
                            newlang.languagename = object.language
                            newlang.isdefault = False
                            newlang.save()
                            allowed_languages = models.DLanguages.objects.values_list('pk', flat=True)

                        relation_or_value_type = predicate.replace(SKOS, '') # this is essentially the skos element type within a <skos:Concept> element (eg: prefLabel, broader, etc...)

                        if relation_or_value_type in skos_value_types_list:
                            value_type = skos_value_types.get(valuetype=relation_or_value_type)
                            concept.addvalue({'value':object, 'language': object.language, 'type': value_type.valuetype, 'category': value_type.category}) 
                        elif predicate == SKOS.broader:
                            self.relations.append({'source': self.generate_uuid_from_subject(baseuuid, object), 'type': 'narrower', 'target': self.generate_uuid_from_subject(baseuuid, s)})
                        elif predicate == SKOS.narrower:
                            self.relations.append({'source': self.generate_uuid_from_subject(baseuuid, s), 'type': relation_or_value_type, 'target': self.generate_uuid_from_subject(baseuuid, object)})
                        elif predicate == SKOS.related:
                            self.relations.append({'source': self.generate_uuid_from_subject(baseuuid, s), 'type': relation_or_value_type, 'target': self.generate_uuid_from_subject(baseuuid, object)})

                self.nodes.append(concept)


            # insert and index the concpets
            with transaction.atomic():
                for node in self.nodes:
                    node.save()
                    node.index(scheme=scheme_id)            

                # insert the concept relations
                for relation in self.relations:
                    newrelation = models.ConceptRelations()
                    newrelation.relationid = str(uuid.uuid4())
                    newrelation.conceptidfrom_id = relation['source']
                    newrelation.conceptidto_id = relation['target']
                    newrelation.relationtype_id = relation['type']
                    newrelation.save()

            return self
        else:
            raise Exception('graph argument should be of type rdflib.graph.Graph')

    def generate_uuid_from_subject(self, baseuuid, subject):
        return str(uuid.uuid3(baseuuid, str(subject)))


class SKOSWriter(object):
    def __init__(self):
        self.concept_list = []

    def write(self, concept_graph, format='pretty-xml'):
        #get empty RDF graph
        rdf_graph = Graph()

        #define namespaces
        ARCHES = Namespace('http://www.archesproject.org/')

        #bind the namespaces
        rdf_graph.bind('arches',ARCHES)
        rdf_graph.bind('skos',SKOS)
        
        """
        #add main concept to the graph
        rdf_graph.add((subject, predicate, object))
        rdf_graph.add((ARCHES[node.id], RDF['type'], SKOS.Concept))
        rdf_graph.add((Arches guid, SKOS.prefLabel, Literal('Stone',lang=en)))
        """

        def build_skos(node):
            conceptid = node.id
            rdf_graph.add((ARCHES[node.id], RDF.type, SKOS[node.nodetype]))

            for value in node.values:
                if value.category == 'label' or value.category == 'note':
                    rdf_graph.add((ARCHES[node.id], SKOS[value.type], Literal(value.value, lang = value.language)))
                else:
                    rdf_graph.add((ARCHES[node.id], ARCHES[value.type.replace(' ', '_')], Literal(value.value, lang = value.language)))

            for subconcept in node.subconcepts:
                if node.nodetype == 'ConceptScheme':
                    rdf_graph.add((ARCHES[node.id], SKOS.hasTopConcept, ARCHES[subconcept.id]))
                elif node.nodetype == 'Concept':
                    rdf_graph.add((ARCHES[node.id], SKOS.narrower, ARCHES[subconcept.id]))

            for relatedconcept in node.relatedconcepts:
                rdf_graph.add((ARCHES[node.id], SKOS.related, ARCHES[relatedconcept.id]))


        concept_graph.traverse(build_skos)
        return rdf_graph.serialize(format=format)


def dictfetchall(cursor):
    "Returns all rows from a cursor as a dict"
    desc = cursor.description
    return [
        dict(zip([col[0] for col in desc], row))
        for row in cursor.fetchall()
    ]


import datetime
import decimal
import types
import json
import inspect
from io import StringIO
from django.db import models, DEFAULT_DB_ALIAS
from django.db.models import Model
from django.db.models.query import QuerySet
from django.utils.encoding import smart_unicode
from django.core.serializers.python import Serializer as PythonSerializer
from django.core.serializers.python import Deserializer as PythonDeserializer
from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.gis.geos import GEOSGeometry
from django.core.files import File

class UnableToSerializeError(Exception):
    """ Error for not implemented classes """
    def __init__(self, value):
        self.value = value
        Exception.__init__(self)

    def __str__(self):
        return repr(self.value)

class UnableToSerializeMethodTypesError(Exception):
    """ Error for not implemented classes """
    def __init__(self, value):
        self.value = value
        Exception.__init__(self)

    def __str__(self):
        return repr(self.value)

class JSONSerializer(object):
    def serializeToPython(self, obj, **options):
        self.options = options.copy()

        self.stream = options.pop("stream", StringIO())
        self.selected_fields = options.pop("fields", None)
        self.use_natural_keys = options.pop("use_natural_keys", False)
        self.geom_format = options.pop("geom_format", "wkt")

        return self.handle_object(obj)

    def serialize(self, obj, **options):
        self.options = options.copy()

        self.stream = options.pop("stream", StringIO())
        self.selected_fields = options.pop("fields", None)
        self.use_natural_keys = options.pop("use_natural_keys", False)
        self.geom_format = options.pop("geom_format", "wkt")

        ret = self.handle_object(obj)

        return json.dumps(ret, cls=DjangoJSONEncoder, **options.copy())


    def handle_object(self, object):
        """ Called to handle everything, looks for the correct handling """
        # print type(object)
        # print inspect.isclass(object)
        # print inspect.ismethod(object)
        # print inspect.isfunction(object)
        # print inspect.isbuiltin(object)
        # print inspect.isroutine(object)
        # print inspect.isabstract(object)
        # print type(object) == 'staticmethod'
        if (inspect.isroutine(object) or
            inspect.isbuiltin(object) or
            inspect.isclass(object)):
            raise UnableToSerializeMethodTypesError(type(object))
        elif isinstance(object, dict):
            return self.handle_dictionary(object)
        elif (isinstance(object, list) or 
              isinstance(object, tuple)):
            return self.handle_list(object)
        elif isinstance(object, Model):
            #return super(JSONSerializer,self).serialize([object], **self.options.copy())[0]
            return PythonSerializer().serialize([object],**self.options.copy())
        elif isinstance(object, QuerySet):
            #return super(JSONSerializer,self).serialize(object, **self.options.copy())[0]
            return PythonSerializer().serialize(object,**self.options.copy())
        elif (isinstance(object, int) or
              isinstance(object, float) or 
              isinstance(object, long) or 
              isinstance(object, basestring) or 
              isinstance(object, bool) or 
              object is None):
            return object
        elif (isinstance(object, datetime.datetime) or
              isinstance(object, datetime.date) or
              isinstance(object, datetime.time) or
              isinstance(object, decimal.Decimal)):
            return DjangoJSONEncoder().default(object)
        elif isinstance(object, GEOSGeometry):
            return getattr(object, self.geom_format)
        elif isinstance(object, File):
            return object.name
        elif hasattr(object, '__dict__'):
            return self.handle_dictionary(object.__dict__)
        else:
            raise UnableToSerializeError(type(object))

    def handle_dictionary(self, d):
        """Called to handle a Dictionary"""
        obj = {}
        for key, value in d.iteritems():
            try:
                #print key + ': ' + str(type(value))
                obj[key] = self.handle_object(value)
            except(UnableToSerializeMethodTypesError):
                pass

        return obj


    def handle_list(self, l):
        """Called to handle a list"""
        arr = []
        for item in l:
            arr.append(self.handle_object(item))

        return arr


class JSONDeserializer(object):
    """
    Deserialize a stream or string of JSON data.
    """
    def deserialize(self, stream_or_string, **options):
        self.options = options.copy()

        self.stream = options.pop("stream", StringIO())
        self.selected_fields = options.pop("fields", None)
        self.use_natural_keys = options.pop("use_natural_keys", False)

        if isinstance(stream_or_string, basestring):
            stream = StringIO(smart_unicode(stream_or_string))
        else:
            stream = stream_or_string

        ret = self.handle_object(json.load(stream))

        return ret


    def handle_object(self, object):
        """ Called to handle everything, looks for the correct handling """
        if isinstance(object, dict):
            if ('pk' in object and 'model' in object and 'fields' in object):
                # assume that this is a serialized django model
                return self.handle_model(object)
            else:
                return self.handle_dictionary(object)
        elif (isinstance(object, list) or 
              isinstance(object, tuple)):
            return self.handle_list(object)
        elif (isinstance(object, int) or
              isinstance(object, float) or 
              isinstance(object, long) or 
              isinstance(object, basestring) or 
              isinstance(object, bool) or 
              object is None):
            return object
        #elif isinstance(object, tuple):
        #    return tuple(self.serialize([item for item in obj]))
        elif hasattr(object, '__dict__'):
            return self.handle_dictionary(object.__dict__)
        else:
            raise UnableToSerializeError(type(object))


    def handle_dictionary(self, d):
        """Called to handle a Dictionary"""
        obj = {}
        for key, value in d.iteritems():
            obj[key] = self.handle_object(value)

        return obj


    def handle_list(self, l):
        """Called to handle a list"""
        arr = []
        for item in l:
            arr.append(self.handle_object(item))

        return arr


    def handle_model(self, m):
        """Called to handle a model"""
        a = []
        for obj in PythonDeserializer([m],**self.options.copy()):
            a.append(obj)
        return a



from django.db import connection, transaction

def insert_concept_relations(sourcelegacyid, relationshiptype, targetlegacyid):
    cursor = connection.cursor()

    sql = """
        SELECT concepts.insert_relation('%s','%s','%s');
    """ % (sourcelegacyid, relationshiptype.replace("'","''"), targetlegacyid)

    # print sql
    cursor.execute(sql)

def index_entity_concept_lables():
    from arches.app.models import models
    from arches.app.models.concept import Concept
    
    domains_concept = Concept('00000000-0000-0000-0000-000000000003')
    entitynodes_concept = Concept('00000000-0000-0000-0000-000000000004')
    
    entitytypes = models.EntityTypes.objects.all()
    for entitytype in entitytypes:
        concept = Concept(entitytype.conceptid).get(include=['label'])
        if entitytype.businesstablename == 'domains':
            concept.index(scheme=domains_concept)
        else:
            concept.index(scheme=entitynodes_concept)



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

import uuid
import types
import copy
import arches.app.models.models as archesmodels
from django.conf import settings
from django.contrib.gis.db import models
from django.contrib.gis.geos import fromstr
from django.db import connection
from django.db import transaction
from django.core.files.uploadedfile import InMemoryUploadedFile, TemporaryUploadedFile
from arches.app.search.search_engine_factory import SearchEngineFactory
from arches.app.utils.betterJSONSerializer import JSONSerializer, JSONDeserializer


class Entity(object):
    """ 
    Used for mapping complete entity graph objects to and from the database

    """
    
    def __init__(self, *args, **kwargs):
        self.property = ''
        self.entitytypeid = ''
        self.entityid = ''
        self.value = ''
        self.label = ''
        self.businesstablename = ''
        self.child_entities = [
            # contains an array of other entities
        ]      

        if len(args) != 0:
            if isinstance(args[0], basestring):
                try:
                    uuid.UUID(args[0])
                    self.get(args[0])
                except(ValueError):
                    self.load(JSONDeserializer().deserialize(args[0]))  
            elif isinstance(args[0], Entity):
                self = args[0]
            elif args[0] != None and isinstance(args[0], object):
                self.load(args[0])  

    def __repr__(self):
        return ('%s: %s of type %s with value "%s"') % (self.__class__, self.entityid, self.entitytypeid, self.value)

    def __hash__(self): return hash(JSONSerializer().serialize(self))
    def __eq__(self, x):return hash(self) == hash(x)
    def __ne__(self, x): return hash(self) != hash(x)

    def get(self, pk, parent=None):
        """
        Gets a complete entity graph for a single entity instance given an entity id
        If a parent is given, will attempt to lookup the rule used to relate parent to child

        """

        entity = archesmodels.Entities.objects.get(pk = pk)
        self.entitytypeid = entity.entitytypeid_id
        self.entityid = entity.pk
        self.businesstablename = entity.entitytypeid.businesstablename if entity.entitytypeid.businesstablename else ''

        # get the entity value if any
        if entity.entitytypeid.businesstablename != None:
            themodel = self._get_model(entity.entitytypeid.businesstablename)
            themodelinstance = themodel.objects.get(pk = pk)
            columnname = entity.entitytypeid.getcolumnname()

            if (isinstance(themodelinstance, archesmodels.Domains)): 
                self.value = themodelinstance.getlabelid()
                self.label = themodelinstance.getlabelvalue()
            elif (isinstance(themodelinstance, archesmodels.Files)): 
                self.label = themodelinstance.getname()
                self.value = themodelinstance.geturl() 
            else:
                self.value = getattr(themodelinstance, columnname, 'Entity %s could not be found in the table %s' % (pk, entity.entitytypeid.businesstablename))                   
                self.label = self.value
                
        # get the property that associated parent to child
        if parent is not None:
            relation = archesmodels.Relations.objects.get(entityiddomain =  parent.entityid, entityidrange = entity.entityid)
            self.property = relation.ruleid.propertyid_id                
        
        # get the child entities if any
        child_entities = archesmodels.Relations.objects.filter(entityiddomain = pk)
        for child_entity in child_entities:       
            self.append_child(Entity().get(child_entity.entityidrange_id, entity))

        return self

    def _save(self):
        """
        Saves an entity back to the db, returns a DB model instance, not an instance of self

        """
        entitytype = archesmodels.EntityTypes.objects.get(pk = self.entitytypeid)
        try:
            uuid.UUID(self.entityid)
        except(ValueError):
            self.entityid = str(uuid.uuid4())


        entity = archesmodels.Entities()
        entity.entitytypeid = entitytype
        entity.entityid = self.entityid
        entity.save()


        columnname = entity.entitytypeid.getcolumnname()
        if columnname != None:
            themodel = self._get_model(entity.entitytypeid.businesstablename)
            themodelinstance = themodel()
            themodelinstance.entityid = entity

            if (isinstance(themodelinstance, archesmodels.Files)): 
                # Saving of files must be handled specially
                # Because on subsequent saves of a file resource, we post back the file path url (instead of posting the file like we originally did),
                # we want to prevent the path from being saved back to the database thus screwing up the file save process 
                if isinstance(self.value, (InMemoryUploadedFile, TemporaryUploadedFile)):
                    setattr(themodelinstance, columnname, self.value)
                    themodelinstance.save()
                elif isinstance(self.value, str):
                    setattr(themodelinstance, columnname, self.value)
                    themodelinstance.save()
            else:
                setattr(themodelinstance, columnname, self.value)
                themodelinstance.save()

            self.businesstablename = entity.entitytypeid.businesstablename
            self.label = self.value
            if (isinstance(themodelinstance, archesmodels.Domains)): 
                self.value = themodelinstance.getlabelid()
                self.label = themodelinstance.getlabelvalue()
            elif (isinstance(themodelinstance, archesmodels.Files)): 
                self.value = themodelinstance.geturl()
                self.label = themodelinstance.getname()

        for child_entity in self.child_entities:
            child = child_entity._save()
            try:
                rule = archesmodels.Rules.objects.get(entitytypedomain = entity.entitytypeid, entitytyperange = child.entitytypeid, propertyid = child_entity.property)
                newrelationship = archesmodels.Relations()
                newrelationship.entityiddomain = entity
                newrelationship.entityidrange = child
                newrelationship.ruleid = rule
                newrelationship.save()
            except:
                print 'ERROR in query for the following rule: Domain={0}, Range={1}, Property={2}. Relationship could not be saved'.format(entity.entitytypeid, child.entitytypeid, entity.property)

        return entity

    def _delete(self, delete_root=False):
        """
        Deletes this entity and all it's children.  
        Also attempts to delete the highest parent (and any nodes on the way) of this node when I'm the only child and 
        my parent has no value.

        if delete_root is False prevent the root node from deleted

        """

        nodes_to_delete = []

        # gather together a list of all entities that includes self and all its children
        def gather_entities(entity):
            nodes_to_delete.append(entity)
        self.traverse(gather_entities)
        
        # delete the remaining entities
        for entity in nodes_to_delete:
            self_is_root = entity.get_rank() == 0

            if self_is_root and delete_root:
                dbentity = archesmodels.Entities.objects.get(pk = entity.entityid)
                #print 'deleting root: %s' % dbentity
                dbentity.delete()
            else:
                parent = entity.get_parent()
                parent_is_root = parent.get_rank() == 0              
                # print 'deleting: %s' % JSONSerializer().serializeToPython(entity, ensure_ascii=True, indent=4)
                dbentity = archesmodels.Entities.objects.filter(pk = entity.entityid)
                if len(dbentity) == 1:
                    dbentity[0].delete()
                    parent.child_entities.remove(entity)
                    # print 'deleted: %s' % dbentity[0]
                    # print 'parent: %s' % JSONSerializer().serializeToPython(parent, ensure_ascii=True, indent=4)
            
                # now try and remove this entity's parent 
                if len(parent.child_entities) == 0 and parent.value == '' and not parent_is_root:
                    #print 'trying to delete parent node'
                    parent._delete()  

    def load(self, E):
        """
        Populate an Entity instance from a generic python object 

        """

        self.property = E.get('property', '')
        self.entitytypeid = E.get('entitytypeid', '')
        self.entityid = E.get('entityid', '')
        self.value = E.get('value', '')
        self.label = E.get('label', '')
        self.businesstablename = E.get('businesstablename', '')
        for entity in E.get('child_entities', []):
            child_entity = Entity()
            self.append_child(child_entity.load(entity))
        return self

    def add_child_entity(self, entitytypeid, property, value, entityid):
        """
        Add a child entity to this entity instance

        """     

        node = Entity()
        node.property = property
        node.entitytypeid = entitytypeid
        node.value = value
        node.entityid = entityid
        # if entitytypeid == 'PLACE.E53': print 'in add related entity'
        self.append_child(node)
        return node

    def append_child(self, entity):
        """
        Append a child entity to this entity instance

        """
        parent = self        
        # if self.entitytypeid == 'PLACE.E53': print 'assigning parent:id:%s -- %s' % (id(self), self)
        # if parent.entitytypeid == 'PLACE.E53': print 'assigning parent2:id:%s -- %s' % (id(parent), parent)
        def func(self):
            # if parent.entitytypeid == 'PLACE.E53': print 'get parent:id:%s -- %s' % (id(parent), parent)
            return parent

        entity.get_parent = types.MethodType(func, entity, Entity)
        if entity not in self.child_entities:
            self.child_entities.append(entity)

    def equals(self, entitytocompare, strict=False):
        """
        Test to compare if 2 entities are equal, if strict=True then compare node values as well

        """
        if (self.entitytypeid == entitytocompare.entitytypeid and self.entityid == entitytocompare.entityid and self.property == entitytocompare.property):
            if strict:
                if self.value == entitytocompare.value:
                    return True
            else:
                return True
        return False

    def merge(self, entitytomerge):
        """
        Merge an entity graph into this instance at the lowest common node

        """        

        # if the nodes are equal attempt a merge otherwise don't bother
        if (self.entitytypeid == entitytomerge.entitytypeid and self.property == entitytomerge.property):
            # if the value of each node is not blank then the nodes can't be merged
            # and we simply append entitytomerge node to self's parent node
            if self.value != '' and entitytomerge.value != '':
                self.get_parent().append_child(entitytomerge)
                return self

            # update self.entityid if it makes sense to do so  
            if self.entityid == '' and entitytomerge.entityid != '':
                self.entityid = entitytomerge.entityid

            # update self.value if it makes sense to do so  
            if self.value == '' and entitytomerge.value != '':
                self.value = entitytomerge.value

            child_entitiesmerged = []
            # try to merge any child_entities of self and entitytomerge 
            for child_entitytomerge in entitytomerge.child_entities:
                for child_entity in self.child_entities:
                    if (child_entity.entitytypeid == child_entitytomerge.entitytypeid and child_entity.property == child_entitytomerge.property):   
                        child_entity.merge(child_entitytomerge)
                        child_entitiesmerged.append(child_entitytomerge)

            # append all entitytomerge.child_entities that weren't merged above
            for child_entity in list(set(entitytomerge.child_entities)-set(child_entitiesmerged)):
                self.append_child(child_entity)
        else:
            self.get_parent().append_child(entitytomerge)
        return self

    def merge_at(self, entitytomerge, entitytypeid):
        """
        Merge an entity graph into this instance at the node type specified
        If the node can't be found in self then merge the entity graph at Root

        """

        selfEntities = self.find_entities_by_type_id(entitytypeid)
        foundEntities = entitytomerge.find_entities_by_type_id(entitytypeid)
        if len(selfEntities) == 1 and len(foundEntities) == 1:
            for foundEntity in foundEntities[0].child_entities:
                selfEntities[0].append_child(foundEntity)
        
        # if you can't find the merge node in self then just merge at Root
        if len(selfEntities) == 0 and len(foundEntities) == 1:
            self.merge_at(entitytomerge, self.entitytypeid)

        return self

    def diff(self, entitytotest):
        """
        Find all the entities in self that don't exist in entitytotest 
        (this represents entities that have effectively been deleted from entitytotest 
        when entitytotest is a version of self)

        """

        ret = {'deleted_nodes':[], 'updated_nodes':[], 'inserted_nodes': []}

        self_flattened = set(self.flatten())
        entitytotest_flattened = set(entitytotest.flatten())

        ret['deleted_nodes'] = list(entitytotest_flattened.difference(self_flattened))
        ret['inserted_nodes'] = list(self_flattened.difference(entitytotest_flattened))

        for inserted_entity in list(self_flattened.difference(entitytotest_flattened)):
            for deleted_entity in list(entitytotest_flattened.difference(self_flattened)):
                if inserted_entity.entityid == deleted_entity.entityid:
                    ret['inserted_nodes'].remove(inserted_entity)
                    ret['deleted_nodes'].remove(deleted_entity)
                    ret['updated_nodes'].append({'from': deleted_entity, 'to': inserted_entity})

        return ret

    def copy(self):
        return copy.deepcopy(self)

    def flatten(self):
        """
        flattens the graph into a list of unordered entities

        """

        ret = []
        def gather_entities(entity):
            if entity.get_rank() != 0:
                entity.parentid = entity.get_parent().entityid
            else:
                entity.parentid = None
            ret.append(entity)

        copiedself = self.copy()
        copiedself.traverse(gather_entities)
        for item in ret:
            item.child_entities = []

        return ret

    def find_entities_by_type_id(self, entitytypeid):
        """
        Gets a list of entities within this instance of a given type

        """
        ret = []
        def appendValue(entity):
            if entity.entitytypeid == entitytypeid:
                ret.append(entity)

        self.traverse(appendValue)
        return ret

    def traverse(self, func, scope=None):
        """
        Traverses a graph from leaf to root calling the given function on each node
        passes an optional scope to each function

        Return a value from the function to prematurely end the traversal

        """

        if scope == None:
            ret = func(self)
        else:
            ret = func(self, scope) 

        # break out of the traversal if the function returns a value
        if ret != None:
            return ret   

        for child_entity in self.child_entities:
            ret = child_entity.traverse(func, scope) 
            if ret != None: 
                return ret   

    def get_rank(self, rank=0):
        """
        Get the rank of this instance (root is 0)

        """
        if hasattr(self, 'get_parent'):
            return self.get_parent().get_rank(rank+1)
        return rank

    def get_root(self):
        """
        Get the root node of this instance

        """
        if hasattr(self, 'get_parent'):
            return self.get_parent().get_root()
        return self

    def set_entity_value(self, entitytypeid, value, append=False):
        """

        """

        entities = self.find_entities_by_type_id(entitytypeid)

        if append or len(entities) == 0:
            schema = Entity.get_mapping_schema(self.entitytypeid)
            entity = Entity()
            entity.create_from_mapping(self.entitytypeid, schema[entitytypeid]['steps'], entitytypeid, value)
            self.merge_at(entity, schema[entitytypeid]['mergenodeid'])
            return entity

        if len(entities) == 1:
            entities[0].value = value 
            return entities[0]

    def create_from_mapping(self, entitytypeid, mappingsteps, leafentitytypeid, leafvalue, leafentityid=''):
        currentEntity = self
        currentEntity.entitytypeid = entitytypeid
        for step in mappingsteps:
            currentEntity.entityid = ''
            value = ''
            if step['entitytyperange'] == leafentitytypeid:
                value = leafvalue
            currentEntity = currentEntity.add_child_entity(step['entitytyperange'], step['propertyid'], value, leafentityid)
        return self

    @classmethod
    def get_mapping_schema(cls, entitytypeid):
        """
        Gets a complete entity schema graph for a single entity type given an entity type id

        """

        ret = {}
        mappings = archesmodels.Mappings.objects.filter(entitytypeidfrom = entitytypeid)

        for mapping in mappings:
            if mapping.entitytypeidto.pk not in ret:
                ret[mapping.entitytypeidto.pk] = {'steps':[], 'mergenodeid': mapping.mergenodeid}

            ret[mapping.entitytypeidto.pk]['steps'] = (Entity._get_mappings(mapping.pk))

        return ret

    @classmethod
    def _get_mappings(cls, mappingid):
        """
        Gets a single mapping given an mapping id

        """

        ret = []
        cursor = connection.cursor()
        cursor.execute("""
            SELECT 
              rules.entitytypedomain, 
              rules.entitytyperange, 
              rules.propertyid, 
              mapping_steps."order", 
              mappings.entitytypeidfrom, 
              mappings.entitytypeidto, 
              mapping_steps.mappingid
            FROM 
              ontology.rules, 
              ontology.mapping_steps, 
              ontology.mappings
            WHERE 
              mapping_steps.ruleid = rules.ruleid AND
              mappings.mappingid = '%s' AND
              mappings.mappingid = mapping_steps.mappingid
            ORDER BY
              mappings.entitytypeidfrom ASC, 
              mappings.entitytypeidto ASC, 
              mappings.mappingid ASC, 
              mapping_steps."order" ASC;
        """ % (mappingid))
        mapping_steps = cursor.fetchall()

        for mapping_step in mapping_steps:
            rule = {}
            rule['entitytypedomain'] = mapping_step[0]
            rule['entitytyperange'] = mapping_step[1]
            rule['propertyid'] = mapping_step[2]

            ret.append(rule)

        return ret


    def prune(self, entitytypes, action='disallow'):
        """
        entitytypes is a list of entitytypeids allowed or dissallowed in the graph

        if action=disallow (the default) then prune will remove all entities and their children from the entity graph that match the list of provided entitytypes
        if action=allow then prune will pass through all entities and their parents from the entity graph that match the list of provided entitytypes

        .. code-block:: python

            # simple example of prunning a graph
            entity.prune(['ARCHES RECORD.E31', 'PHASE TYPE ASSIGNMENT.E17', 'SPATIAL COORDINATES_GEOMETRY.E47', 'NAME.E41'])
            
            # a more fully formed example of prunning a graph based on user permissions
            fullgraph = set(entity.flatten())
            from django.contrib.auth.models import User
            user = User.objects.get(pk=1)
            print user
            if not user.is_superuser:
                permissions = user.get_all_permissions()
                entitytypes = []
                for permission in permissions:
                    if permission.startswith('%s.read_' % entity.entitytypeid):
                        print permission.split('%s.read_' % entity.entitytypeid)[1]
                        entitytypes.append(permission.split('%s.read_' % entity.entitytypeid)[1])

                entity.prune(entitytypes, action='allow')
                prunedgraph = set(entity.flatten())
                print fullgraph.intersection(prunedgraph)
                print fullgraph.issuperset(prunedgraph)
                print fullgraph.symmetric_difference(prunedgraph)

        """

        parent_entitytypes = set()
        flattened_graph = self.flatten()
        entities_to_prune = set()        

        def gather_parent_entitytypes(entity):
            if entity.get_rank() == 0:
                return
            parent_entity = entity.get_parent()
            parent_entitytypes.add(parent_entity.entitytypeid)
            gather_parent_entitytypes(parent_entity)

        if action == 'disallow':
            for entity in flattened_graph:
                if entity.entitytypeid in entitytypes:
                    entities_to_prune.add(entity)

        else:
            # if you passed in no entitytypes then you're basically saying remove all information from the graph
            if len(entitytypes) == 0:
                self.clear()
                return

            # first we need to loop through the graph to all the parents of entity to the list of allowed entitytypes
            for entity in flattened_graph:
                if entity.entitytypeid in entitytypes:
                    gather_parent_entitytypes(entity)

            entitytypes = entitytypes + list(parent_entitytypes)
            for entity in flattened_graph:
                if entity.entitytypeid not in entitytypes:
                    entities_to_prune.add(entity)
    
        # prune the remaining entities
        print 'entities to prune: %s' % entities_to_prune
        for entity in entities_to_prune:
            try:
                parent = entity.get_parent()   
                print '\nremoving: %s' % entity         
                parent.child_entities.remove(entity) 
            except:
                if entity.get_rank() == 0:
                    self.clear()
                    return

        self.trim()

    def trim(self):
        """
        recusrsively removes all nodes starting from the leaf that have no child_entities and no value
        these nodes are assumed to be of no material value to the graph

        """

        def func(entity):
            try:
                parent = entity.get_parent()
                if len(entity.child_entities) == 0 and entity.value == '':
                    parent.child_entities.remove(entity)
            except:
                pass
        self.traverse(func)

    def clear(self):
        """
        resets this entity back to a clean state (does not delete the entity from the database)

        """

        self.child_entities = []
        self.entitytypeid = ''
        self.entityid = ''
        self.value = ''

    @staticmethod
    def _get_model(tablename):
        """
        Helper to look up a model from a table name.

        """

        try:
            model_identifier = str('models.' + tablename)
            Model = models.get_model(*model_identifier.split("."))
        except TypeError:
            Model = None
        if Model is None:
            raise TypeError(u"Invalid model identifier: '%s'" % model_identifier)
        return Model


    def dictify(self, keys=['label']):
        """
        Takes an entity and turns it into recursive lists nested objects
        Uses an in-built algorithm to derive which sub-branches appear to be grouped, and then flattens them out

        A partial example output follows...

        .. code-block:: python
            [
                {
                    "EVALUATION_CRITERIA_ASSIGNMENT_E13": [{
                        "STATUS_E55": [
                            {"STATUS_E55__label": "3D"}, 
                            {"STATUS_E55__label": "3CD"}, 
                            {"STATUS_E55__label": "5D3"}
                        ]
                    }], 
                    "BEGINNING_OF_EXISTENCE_E63": [{
                        "BEGINNING_OF_EXISTENCE_TIME-SPAN_E52": [{
                            "BEGINNING_OF_EXISTENCE_TIME-SPAN_E52__label": "", 
                            "START_DATE_OF_EXISTENCE_E49__label": "1962-01-01T00:00:00"
                        }], 
                        "BEGINNING_OF_EXISTENCE_TYPE_E55": [{
                            "BEGINNING_OF_EXISTENCE_TYPE_E55__label": "Built Date"
                        }]
                    }], 
                    "NAME_E41": [{
                        "NAME_TYPE_E55__label": "Primary", 
                        "NAME_E41__label": "3264 N WRIGHTWOOD DR"
                    }], 
                    "PRODUCTION_E12": [{
                        "PHASE_TYPE_ASSIGNMENT_E17": [
                            {
                                "STYLE_E55": [{
                                        "STYLE_E55__label": "Modern, Mid-Century"
                                }], 
                                "HERITAGE_RESOURCE_TYPE_E55": [{
                                    "HERITAGE_RESOURCE_TYPE_E55__label": "HP02. Single family property"
                                    },{
                                    "HERITAGE_RESOURCE_TYPE_E55__label": "House"
                                }], 
                                "HERITAGE_RESOURCE_USE_TYPE_E55": [{
                                    "HERITAGE_RESOURCE_USE_TYPE_E55__label": "Historic"
                                }]
                            }
                        ]
                    }]
                }
            ]

        """

        data = {}
        for child_entity in self.child_entities:
            if child_entity.businesstablename != '':
                data[child_entity.undotify()] = self.get_nodes(child_entity.entitytypeid, keys=keys)
            else:
                data[child_entity.undotify()] = child_entity.dictify(keys=keys)
        return [data]

    def get_nodes(self, entitytypeid, keys=[]):
        """
        Used by dictify to gather and flatten a single node (by entitytypeid) and all it's children

        for example, a NAME.E41 node with a single child of NAME_TYPE.E55 would be transformed as below
        
        .. code-block:: python

                "NAME_E41": [{
                    "NAME_TYPE_E55__label": "Primary", 
                    "NAME_E41__label": "3264 N WRIGHTWOOD DR"
                }],

        """

        ret = []
        entities = self.find_entities_by_type_id(entitytypeid)
        for entity in entities:
            data = {}

            for entity in entity.flatten():
                data = dict(data.items() + entity.encode(keys=keys).items())
            ret.append(data)
        return ret

    def encode(self, keys=[]):
        """
        Encodes an Entity into a dictionary of keys derived by the entitytypeid of the Entity concatonated wtih property name 

        .. code-block:: python

                {
                    "NAME_TYPE_E55__label": "Primary"
                }

        """

        ret = {}

        for key, value in self.__dict__.items():
            if key in keys:
                ret['%s__%s' % (self.undotify(), key)] = value
        return ret

    def undotify(self):
        return self.undotify_entitytypeid()

    def undotify_entitytypeid(self):
        return self.entitytypeid.replace('.', '_').replace('-', '___');


import sys
import os
import subprocess
import shutil
import urllib2
import zipfile
import datetime
from arches import settings

here = os.path.dirname(os.path.abspath(__file__))
root_dir = os.path.dirname(here)

def install():
    if confirm_system_requirements():

        run_virtual_environment()

        install_dir = os.path.join(site_packages_dir(), 'arches', 'install')
        django_install_location = os.path.join(site_packages_dir(), 'django')

        # INSTALL DJANGO, RAWES, SPHINX AND OTHER DEPENDENCIES
        tmpinstalldir = os.path.join(site_packages_dir(), 'arches', 'tmp')
        os.system("pip install -b %s setuptools --no-use-wheel --upgrade" % (tmpinstalldir))
        os.system("pip install -b %s -r %s" % (tmpinstalldir, os.path.join(install_dir, 'requirements.txt')))
        if settings.MODE == 'DEV':
            os.system("pip install -b %s -r %s" % (tmpinstalldir, os.path.join(install_dir, 'requirements_dev.txt')))
        shutil.rmtree(tmpinstalldir, True)

        #INSTALLING CUSTOM DJANGO EDITS/PATCHES
        shutil.copy2(os.path.join(install_dir, 'django_overrides', 'base.py'), os.path.join(django_install_location, 'db', 'backends', 'postgresql_psycopg2'))
        shutil.copy2(os.path.join(install_dir, 'django_overrides', 'creation.py'), os.path.join(django_install_location, 'db', 'backends', 'postgresql_psycopg2'))
        shutil.copy2(os.path.join(install_dir, 'django_overrides', 'inspectdb.py'), os.path.join(django_install_location, 'core', 'management', 'commands'))
        shutil.copy2(os.path.join(install_dir, 'django_overrides', 'admin.py'), os.path.join(django_install_location, 'contrib', 'auth'))
        shutil.copy2(os.path.join(install_dir, 'django_overrides', 'models.py'), os.path.join(django_install_location, 'contrib', 'auth'))
        shutil.copy2(os.path.join(install_dir, 'django_overrides', 'widgets.css'), os.path.join(django_install_location, 'contrib', 'admin', 'static', 'admin', 'css'))

        # GET ELASTICSEARCH
        download_elasticsearch(install_dir)

        # INSTALL PSYCOPG2
        if sys.platform == 'win32':
            is_64bit_python = sys.maxsize > 2**32
            if os.path.exists('C:\Program Files (x86)') and is_64bit_python:
                os.system("easy_install http://www.stickpeople.com/projects/python/win-psycopg/2.5.4/psycopg2-2.5.4.win-amd64-py2.7-pg9.3.5-release.exe")
            else:
                os.system("easy_install http://www.stickpeople.com/projects/python/win-psycopg/2.5.4/psycopg2-2.5.4.win32-py2.7-pg9.3.5-release.exe")
        else:
            # Install psycopg2 through pip - Works fine if the correct header files are present
            # See http://goshawknest.wordpress.com/2011/02/16/how-to-install-psycopg2-under-virtualenv/
            os.system("pip install psycopg2==2.5.4")

def site_packages_dir():
    if sys.platform == 'win32':
        return os.path.join(sys.prefix, 'Lib', 'site-packages')
    else:
        py_version = 'python%s.%s' % (sys.version_info[0], sys.version_info[1])
        return os.path.join(sys.prefix, 'lib', py_version, 'site-packages')        

def confirm_system_requirements():
    # CHECK PYTHON VERSION
    if sys.version_info < (2, 7) or sys.version_info >= (3, 0):
        print('ERROR: Arches requires Python 2.7.x')
        sys.exit(101)
    else:
        pass

    # CHECK POSTGRES VERSION
    try:
        postgres_version = subprocess.check_output(["psql", "--version"])
    except OSError:
        print('ERROR: Arches requires psql. Please install and then rerun this file again.')
        sys.exit(101)
    if postgres_version.find("9.") == -1:
        print('ERROR: Arches requires Postgres 9.0 or greater')
        print('Version detected: %s\n' % (postgres_version))
        postgres_override = raw_input('Would like to continue anyway?\nPress Y for Yes or N for No:')
        if(postgres_override == 'Y' or postgres_override == 'y'):
            pass
        else:
            sys.exit(101)
    else:
        pass

    return True

def run_virtual_environment(env='ENV'):
    # is the site running in a virtual environment?
    if hasattr(sys, 'real_prefix'):
        pass # we're running in a virtualenv
    else:
        # Are we a developer who has access to the included virtual env?
        # If so, then try to install and activate the virtual env
        virtualenv_root = os.path.join(root_dir, 'virtualenv')            
        if os.path.exists(os.path.join(virtualenv_root, 'virtualenv.py')):
            virtualenv_working_dir = os.path.join(virtualenv_root, env)
            os.system("python %s %s" % (os.path.join(virtualenv_root, 'virtualenv.py'), virtualenv_working_dir))
            activate_env(virtualenv_working_dir)
        else:
            os.system("pip install virtualenv")
            virtualenv_working_dir = os.path.join(here, 'virtualenv', env)
            os.system("virtualenv %s" % (virtualenv_working_dir))
            if os.path.exists(virtualenv_working_dir):
                activate_env(virtualenv_working_dir)
            else:
                raise Exception("""\n
            ----------------------------------------------------------------------------------------------------------------------
                ERROR: Arches has to be run within a virtual environment http://virtualenv.readthedocs.org/
            -----------------------------------------------------------------------------------------------------------------------\n""")

def activate_env(path_to_virtual_env):
    # ACIVATE THE VIRTUAL ENV
    if sys.platform == 'win32':
        activate_this = os.path.join(path_to_virtual_env, 'Scripts', 'activate_this.py')
    else:
        activate_this = os.path.join(path_to_virtual_env, 'bin', 'activate_this.py')
    execfile(activate_this, dict(__file__=activate_this))

# INSTALL ELASTICSEARCH and HEAD plugin
def download_file(url, file_name):
    u = urllib2.urlopen(url)
    f = open(file_name, 'wb')
    meta = u.info()
    file_size = int(meta.getheaders("Content-Length")[0])
    print "Downloading: %s Bytes: %s" % (file_name, file_size)

    file_size_dl = 0
    block_sz = 8192
    while True:
        buffer = u.read(block_sz)
        if not buffer:
            break

        file_size_dl += len(buffer)
        f.write(buffer)
        status = r"%10d  [%3.2f%%]" % (file_size_dl, file_size_dl * 100. / file_size)
        status = status + chr(8)*(len(status)+1)
        print status,

    f.close()

def unzip_file(file_name, unzip_location):
    with zipfile.ZipFile(file_name, 'r') as myzip:
        print 'unzipping %s to: %s' % (file_name, unzip_location)
        myzip.extractall(unzip_location)

def download_elasticsearch(install_dir):
    url = get_elasticsearch_download_url(install_dir)
    file_name = url.split('/')[-1]
    if not os.path.isfile(os.path.join(install_dir, file_name)):
        download_file(url, os.path.join(install_dir, file_name))

def get_elasticsearch_download_url(install_dir):
    with open(os.path.join(install_dir, "requirements.txt"), "r") as f:
        for line in f:
            if line.startswith('# https://'):
                return line.replace('# ', '').strip()
    raise Exception("""\n
------------------------------------------------------------------------------------------------------
    ERROR: There was an error getting the url for Elastic search from the requirements.txt file
    Make sure the requirements.txt file contains a line similar to the following line,\nincluding the pound symbol (#) but not the double quotes (") and where the x.x.x represent the version number:
        "# https://download.elasticsearch.org/elasticsearch/elasticsearch/elasticsearch-x.x.x.zip"
----------------------------------------------------------------------------------------------------\n""") 

def get_version(version=None):
    "Returns a PEP 440-compliant version number from VERSION."
    version = get_complete_version(version)

    # Now build the two parts of the version number:
    # major = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|rc}N - for alpha, beta and rc releases

    major = get_major_version(version)

    sub = ''
    if version[3] == 'alpha' and version[4] == 0:
        changeset = get_changeset()
        if changeset:
            sub = '.dev%s' % changeset

    elif version[3] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'rc'}
        sub = mapping[version[3]] + str(version[4])

    return str(major + sub)


def get_major_version(version=None):
    "Returns major version from VERSION."
    version = get_complete_version(version)
    parts = 2 if version[2] == 0 else 3
    major = '.'.join(str(x) for x in version[:parts])
    return major


def get_complete_version(version=None):
    """Returns a tuple of the django version. If version argument is non-empty,
    then checks for correctness of the tuple provided.
    """
    if version is None:
        from arches import VERSION as version
    else:
        assert len(version) == 5
        assert version[3] in ('alpha', 'beta', 'rc', 'final')

    return version

def get_changeset(path_to_file=None):
    import os
    import subprocess
    from StringIO import StringIO
    from management.commands.utils import write_to_file

    sb = StringIO()
    if not path_to_file:
        path_to_file =os.path.abspath(os.path.dirname(__file__))
    
    ver = ''
    try:
        hg_archival = open(os.path.abspath(os.path.join(here, '..', '.hg_archival.txt')),'r')
        the_file = hg_archival.readlines()
        hg_archival.close()
        node = ''
        latesttag = ''
        date = ''
        for line in the_file:
            if line.startswith('node:'):
                node = line.split(':')[1].strip()[:12]
            if line.startswith('latesttag:'):
                latesttag = line.split(':')[1].strip()
            if line.startswith('date:'):
                date = line.split(':')[1].strip()
    
        sb.writelines(['__VERSION__="%s"' % latesttag])   
        sb.writelines(['\n__BUILD__="%s"' % node])  
        ver = '%s:%s' % (latesttag, node)
        ver = date
        #write_to_file(os.path.join(path_to_file,'version.py'), sb.getvalue(), 'w')
    except:
        try:
            ver = subprocess.check_output(['hg', 'log', '-r', '.', '--template', '{latesttag}:{node|short}'])
            ver = subprocess.check_output(['hg', 'log', '-r', '.', '--template', '{node|short}'])
            ver = subprocess.check_output(['hg', 'log', '-r', '.', '--template', '{date}'])
            sb.writelines(['__VERSION__="%s"' % ver.split(':')[0]])
            sb.writelines(['\n__BUILD__="%s"' % ver.split(':')[1]]) 
            #write_to_file(os.path.join(path_to_file,'version.py'), sb.getvalue(), 'w')
        except:
            pass

    try:
        timestamp = datetime.datetime.utcfromtimestamp(float(ver))
    except ValueError:
        return None
    return timestamp.strftime('%Y%m%d%H%M%S')

if __name__ == "__main__":
    install()


import os
import sys 
import uuid
import traceback
import unicodecsv
import concepts
import string
from time import time
from os import listdir
from os.path import isfile, join, exists
from django.db import connection, transaction
from django.conf import settings
from arches.app.models import models
from arches.app.models.concept import Concept, ConceptValue
from .. import utils

class Lookups(object):
    def __init__(self, *args, **kwargs):
        self.lookup = {}
        self.concept_relationships = []

    def add_lookup(self, concept=None, rownum=None):
        self.lookup[concept.legacyoid] = {'concept': concept, 'rownum': rownum}

    def get_lookup(self, legacyoid=''):
        if legacyoid in self.lookup:
            return self.lookup[legacyoid]['concept']
        else:
            raise Exception('Legacyoid (%s) not found.  Make sure your ParentConceptid in the csv file references a previously saved concept.' % (legacyoid))

    def add_relationship(self, source=None, type=None, target=None, rownum=None):
        self.concept_relationships.append({'source': source, 'type': type, 'target': target, 'rownum': rownum})

def load_authority_files(break_on_error=True):
    cursor = connection.cursor()
    file_list = []
    errors = []
    
    for path in settings.CONCEPT_SCHEME_LOCATIONS:
        if os.path.exists(path):
            print '\nLOADING AUTHORITY FILES (%s)' % (path)
            print '-----------------------'            
            
            for f in listdir(path):
                if isfile(join(path,f)) and '.values.csv' not in f and f != 'ENTITY_TYPE_X_ADOC.csv' and f[-4:] == '.csv':
                    file_list.append(f)
            
            file_list.sort()

            auth_file_to_entity_concept_mapping = entitytype_to_auth_doc_mapping(cursor, path)
            count = 1
            for file_name in file_list:
                errors = errors + load_authority_file(cursor, path, file_name, auth_file_to_entity_concept_mapping)
                if count > 10:
                    pass
                    #break
                count = count + 1
            #errors = errors + create_link_to_entity_types(cursor, path)
        else:
            errors.append('\n\nPath in settings.CONCEPT_SCHEME_LOCATIONS doesn\'t exist (%s)' % (path))  

    utils.write_to_file(os.path.join(settings.PACKAGE_ROOT, 'logs', 'authority_file_errors.txt'), '')
    if len(errors) > 0:
        utils.write_to_file(os.path.join(settings.PACKAGE_ROOT, 'logs', 'authority_file_errors.txt'), '\n'.join(errors))
        print "\n\nERROR: There were errors in some of the authority files."
        print "Please review the errors at %s, \ncorrect the errors and then rerun this script." % (os.path.join(settings.PACKAGE_ROOT, 'logs', 'authority_file_errors.txt'))
        if break_on_error:
            sys.exit(101)


def load_authority_file(cursor, path_to_authority_files, filename, auth_file_to_entity_concept_mapping):
    print filename.upper()    

    start = time()
    value_types = models.ValueTypes.objects.all()
    filepath = os.path.join(path_to_authority_files, filename)
    unicodecsv.field_size_limit(sys.maxint)
    errors = []
    lookups = Lookups()

    #create nodes for each authority document file and relate them to the authority document node in the concept schema
    auth_doc_file_name = str(filename)
    display_file_name = string.capwords(auth_doc_file_name.replace('_',' ').replace('AUTHORITY DOCUMENT.csv', '').strip())
    if auth_doc_file_name.upper() != 'ARCHES RESOURCE CROSS-REFERENCE RELATIONSHIP TYPES.E32.CSV':
        concept = Concept()
        concept.id = str(uuid.uuid4())
        concept.nodetype = 'ConceptScheme'       
        concept.legacyoid = auth_doc_file_name
        concept.addvalue({'value':display_file_name, 'language': 'en-us', 'type': 'prefLabel', 'category': 'label'})
        scheme_id = concept
        lookups.add_relationship(source='00000000-0000-0000-0000-000000000001', type='narrower', target=concept.id)

    else:
        concept = Concept().get(id = '00000000-0000-0000-0000-000000000005')
        concept.legacyoid = 'ARCHES RESOURCE CROSS-REFERENCE RELATIONSHIP TYPES.E32.csv'
        scheme_id = concept

    lookups.add_lookup(concept=concept, rownum=0)
    
    try:
        with open(filepath, 'rU') as f:
            rows = unicodecsv.DictReader(f, fieldnames=['CONCEPTID','PREFLABEL','ALTLABELS','PARENTCONCEPTID','CONCEPTTYPE','PROVIDER'], 
                encoding='utf-8-sig', delimiter=',', restkey='ADDITIONAL', restval='MISSING')
            rows.next() # skip header row
            for row in rows:              
                try:
                    if 'MISSING' in row:
                        raise Exception('The row wasn\'t parsed properly. Missing %s' % (row['MISSING']))
                    else:
                        concept = Concept()
                        concept.id = str(uuid.uuid4())
                        concept.nodetype = 'Concept'# if row[u'CONCEPTTYPE'].upper() == 'INDEX' else 'Collection'
                        concept.legacyoid = row[u'CONCEPTID']
                        concept.addvalue({'value':row[u'PREFLABEL'], 'language': 'en-us', 'type': 'prefLabel', 'category': 'label'})
                        if row['CONCEPTTYPE'].lower() == 'collector':
                            concept.addvalue({'value':row[u'PREFLABEL'], 'language': 'en-us', 'type': 'collector', 'category': 'label'})
                        if row[u'ALTLABELS'] != '':
                            altlabel_list = row[u'ALTLABELS'].split(';')
                            for altlabel in altlabel_list:
                                concept.addvalue({'value':altlabel, 'language': 'en-us', 'type': 'altLabel', 'category': 'label'})    
                        if lookups.get_lookup(row[u'PARENTCONCEPTID']).values[0].type == 'collector':
                            lookups.add_relationship(source=lookups.get_lookup(legacyoid=row[u'PARENTCONCEPTID']).id, type='member', target=concept.id, rownum=rows.line_num)

                        lookups.add_relationship(source=lookups.get_lookup(legacyoid=row[u'PARENTCONCEPTID']).id, type='narrower', target=concept.id, rownum=rows.line_num)
                        
                        if auth_doc_file_name in auth_file_to_entity_concept_mapping:
                            lookups.add_relationship(source=auth_file_to_entity_concept_mapping[auth_doc_file_name]['ENTITYTYPE_CONCEPTID'], type='member', target=concept.id, rownum=rows.line_num)

                        if row[u'PARENTCONCEPTID'] == '' or (row[u'CONCEPTTYPE'].upper() != 'INDEX' and row[u'CONCEPTTYPE'].upper() != 'COLLECTOR'):
                            raise Exception('The row has invalid values.')

                        lookups.add_lookup(concept=concept, rownum=rows.line_num)    
                        
                except Exception as e:
                    errors.append('ERROR in row %s: %s' % (rows.line_num, str(e)))           
    
    except UnicodeDecodeError as e:
        errors.append('ERROR: Make sure the file is saved with UTF-8 encoding\n%s\n%s' % (str(e), traceback.format_exc()))
    except Exception as e:
        errors.append('ERROR: %s\n%s' % (str(e), traceback.format_exc()))
    
    if len(errors) > 0:
        errors.insert(0, 'ERRORS IN FILE: %s\n' % (filename))
        errors.append('\n\n\n\n')

    try:
        # try and open the values file if it exists
        if exists(filepath.replace('.csv', '.values.csv')):
            with open(filepath.replace('.csv', '.values.csv'), 'rU') as f:
                rows = unicodecsv.DictReader(f, fieldnames=['CONCEPTID','VALUE','VALUETYPE','PROVIDER'], 
                    encoding='utf-8-sig', delimiter=',', restkey='ADDITIONAL', restval='MISSING')
                rows.next() # skip header row
                for row in rows:
                    try:
                        if 'ADDITIONAL' in row:
                            raise Exception('The row wasn\'t parsed properly. Additional fields found %s.  Add quotes to values that have commas in them.' % (row['ADDITIONAL']))
                        else:
                            row_valuetype = row[u'VALUETYPE'].strip()
                            if row_valuetype not in value_types.values_list('valuetype', flat=True): 
                                valuetype = models.ValueTypes()
                                valuetype.valuetype = row_valuetype
                                valuetype.category = 'undefined'
                                valuetype.namespace = 'arches'
                                valuetype.save()
                            
                            value_types = models.ValueTypes.objects.all()
                            concept = lookups.get_lookup(legacyoid=row[u'CONCEPTID'])
                            category = value_types.get(valuetype=row_valuetype).category
                            concept.addvalue({'value':row[u'VALUE'], 'type': row[u'VALUETYPE'], 'category': category})

                    except Exception as e:
                        errors.append('ERROR in row %s (%s): %s' % (rows.line_num, str(e), row))
    
    except UnicodeDecodeError as e:
        errors.append('ERROR: Make sure the file is saved with UTF-8 encoding\n%s\n%s' % (str(e), traceback.format_exc()))
    except Exception as e:
        errors.append('ERROR: %s\n%s' % (str(e), traceback.format_exc()))            
        
    if len(errors) > 0:
        errors.insert(0, 'ERRORS IN FILE: %s\n' % (filename.replace('.csv', '.values.csv')))
        errors.append('\n\n\n\n')


    # insert and index the concpets
    for key in lookups.lookup:
        try:
            lookups.lookup[key]['concept'].save()
        except Exception as e:
            errors.append('ERROR in row %s (%s):\n%s\n' % (lookups.lookup[key]['rownum'], str(e), traceback.format_exc()))
        
        lookups.lookup[key]['concept'].index(scheme=scheme_id)            

    # insert the concept relations
    for relation in lookups.concept_relationships:
        sql = """
            INSERT INTO concepts.relations(conceptidfrom, conceptidto, relationtype)
            VALUES ('%s', '%s', '%s');
        """%(relation['source'], relation['target'], relation['type'])
        #print sql
        try:
            cursor.execute(sql)
        except Exception as e:
            errors.append('ERROR in row %s (%s):\n%s\n' % (relation['rownum'], str(e), traceback.format_exc()))
    
    if len(errors) > 0:
        errors.insert(0, 'ERRORS IN FILE: %s\n' % (filename))
        errors.append('\n\n\n\n')

    #print 'Time to parse = %s' % ("{0:.2f}".format(time() - start))    

    return errors

def entitytype_to_auth_doc_mapping(cursor, path_to_authority_files):
    filepath = os.path.join(path_to_authority_files, 'ENTITY_TYPE_X_ADOC.csv')
    errors = []
    ret = {}
    with open(filepath, 'rU') as f:
        rows = unicodecsv.DictReader(f, fieldnames=['ENTITYTYPE','AUTHORITYDOC'], 
                    encoding='utf-8-sig', delimiter=',', restkey='ADDITIONAL', restval='MISSING')
        rows.next() # skip header row
        for row in rows:
            if row[u'ENTITYTYPE'] != 'ARCHES RESOURCE CROSS-REFERENCE RELATIONSHIP TYPES.E32.csv':
                sql = """
                    SELECT conceptid FROM data.entity_types WHERE entitytypeid = '%s'
                """%(row[u'ENTITYTYPE'])
                #print sql

            try:
                cursor.execute(sql)
                entity_type_conceptid = str(cursor.fetchone()[0])
                ret[str(row['AUTHORITYDOC'])] = {'ENTITYTYPE' : row[u'ENTITYTYPE'], 'ENTITYTYPE_CONCEPTID': entity_type_conceptid}
            except Exception as e:
                errors.append('ERROR in row %s (%s):\n%s\n%s' % (rows.line_num, str(e), sql, traceback.format_exc()))

    if len(errors) > 0:
        errors.insert(0, 'ERRORS IN FILE: %s\n' % (filepath))
        errors.append('\n\n\n\n')
    return ret




'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

import uuid
from datetime import datetime
from django.conf import settings
from django.contrib.gis.geos import fromstr
import arches.app.models.models as archesmodels
from django.db.models import Q
from arches.app.models.entity import Entity
from arches.app.models.concept import Concept
from arches.app.search.search_engine_factory import SearchEngineFactory
from arches.app.utils.betterJSONSerializer import JSONSerializer, JSONDeserializer
from django.utils.translation import ugettext as _
from django.forms.models import model_to_dict
from django.contrib.gis.geos import GEOSGeometry

class Resource(Entity):
    """ 
    Used for managing Resource type entities

    This class will return an instance of the class defined in settings.RESOURCE_MODEL
    The class defined in settings.ENTITY_MODEL must be a subclass of this class (arches.app.models.resource.Resource)

    """
    def __new__(cls, *args, **kwargs):
        modulename = kwargs.get('mod', 'default')
        if modulename == '':
            return super(Resource, cls).__new__(cls)
        else:
            fully_qualified_modulename = settings.RESOURCE_MODEL.get(modulename)
            if fully_qualified_modulename == 'arches.app.models.resource.Resource' or fully_qualified_modulename == '':
                return super(Entity, cls).__new__(cls)
            components = fully_qualified_modulename.split('.')
            classname = components[len(components)-1]
            modulename = ('.').join(components[0:len(components)-1])
            kwargs['mod'] = ''
            mod = __import__(modulename, globals(), locals(), [classname], -1)
            if issubclass(getattr(mod, classname), Resource):
                return getattr(mod, classname)(*args, **kwargs)
            else:
                raise Exception('The class "%s" defined in settings.RESOURCE_MODEL must be a subclass of arches.app.models.resource.Resource' % fully_qualified_modulename)


    def __init__(self, *args, **kwargs):
        super(Resource, self).__init__(*args, **kwargs)

        self.form_groups = []

    def save(self, user={}, note=''):
        """
        Saves a resource back to the db 

        """

        newentity = False
        timestamp = datetime.now()

        if self.entityid != '':
            entity_pre_save = Entity().get(self.entityid)
        else:
            newentity = True

        self._save() 

        if not newentity:
            diff = self.diff(entity_pre_save)
            for entity in diff['deleted_nodes']:
                if entity.label != '' or entity.value != '':
                    edit = archesmodels.EditLog()        
                    edit.editlogid = str(uuid.uuid4())
                    edit.resourceentitytypeid = self.entitytypeid
                    edit.resourceid = self.entityid
                    edit.userid = getattr(user, 'id', '')
                    edit.user_email = getattr(user, 'email', '')
                    edit.user_firstname = getattr(user, 'first_name', '')
                    edit.user_lastname = getattr(user, 'last_name', '')
                    edit.note = note
                    edit.timestamp = timestamp
                    edit.attributeentitytypeid = entity.entitytypeid
                    edit.edittype = 'delete'
                    edit.oldvalue = entity.label if entity.label != '' else entity.value
                    edit.save()             
                entity._delete(delete_root=True)

            for entity in diff['updated_nodes']:
                if entity['from'].label != '' or entity['to'].label != '' or entity['from'].value != '' or entity['to'].value != '':
                    edit = archesmodels.EditLog()        
                    edit.editlogid = str(uuid.uuid4())
                    edit.resourceentitytypeid = self.entitytypeid
                    edit.resourceid = self.entityid
                    edit.userid = getattr(user, 'id', '')
                    edit.user_email = getattr(user, 'email', '')
                    edit.user_firstname = getattr(user, 'first_name', '')
                    edit.user_lastname = getattr(user, 'last_name', '')
                    edit.note = note
                    edit.timestamp = timestamp
                    edit.attributeentitytypeid = entity['from'].entitytypeid
                    edit.edittype = 'update'
                    edit.oldvalue = entity['from'].label if entity['from'].label != '' else entity['from'].value 
                    edit.newvalue = entity['to'].label if entity['to'].label != '' else entity['to'].value 
                    edit.save()    

            for entity in diff['inserted_nodes']:
                if entity.label != '' or entity.value != '':
                    edit = archesmodels.EditLog()        
                    edit.editlogid = str(uuid.uuid4())
                    edit.resourceentitytypeid = self.entitytypeid
                    edit.resourceid = self.entityid
                    edit.userid = getattr(user, 'id', '')
                    edit.user_email = getattr(user, 'email', '')
                    edit.user_firstname = getattr(user, 'first_name', '')
                    edit.user_lastname = getattr(user, 'last_name', '')
                    edit.note = note
                    edit.timestamp = timestamp
                    edit.attributeentitytypeid = entity.entitytypeid
                    edit.edittype = 'insert'
                    edit.oldvalue = None
                    edit.newvalue = entity.label if entity.label != '' else entity.value
                    edit.save()      

        else:
            for entity in self.flatten():
                if entity.label != '' or entity.value != '':
                    edit = archesmodels.EditLog()        
                    edit.editlogid = str(uuid.uuid4())
                    edit.resourceentitytypeid = self.entitytypeid
                    edit.resourceid = self.entityid
                    edit.userid = getattr(user, 'id', '')
                    edit.user_email = getattr(user, 'email', '')
                    edit.user_firstname = getattr(user, 'first_name', '')
                    edit.user_lastname = getattr(user, 'last_name', '')
                    edit.note = note
                    edit.timestamp = timestamp
                    edit.attributeentitytypeid = entity.entitytypeid
                    edit.edittype = 'create'
                    edit.oldvalue = None
                    edit.newvalue = entity.label if entity.label != '' else entity.value
                    edit.save()

        return self

    def delete(self, user={}, note=''):
        """
        Deltes a resource from the db
        
        """

        timestamp = datetime.now()
        for entity in self.flatten():
            edit = archesmodels.EditLog()        
            edit.editlogid = str(uuid.uuid4())
            edit.resourceentitytypeid = self.entitytypeid
            edit.resourceid = self.entityid
            edit.userid = getattr(user, 'id', '')
            edit.user_email = getattr(user, 'email', '')
            edit.user_firstname = getattr(user, 'first_name', '')
            edit.user_lastname = getattr(user, 'last_name', '') 
            edit.note = note                               
            edit.timestamp = timestamp
            edit.attributeentitytypeid = entity.entitytypeid
            edit.edittype = 'delete'
            edit.oldvalue = entity.label if entity.label != '' else entity.value
            edit.newvalue = None
            edit.save()

        self._delete(delete_root=True)

    def get_form(self, form_id):
        selected_form = None
        forms = [form for group in self.form_groups for form in group['forms']]
        for form in forms:
            if form['id'] == form_id:
                selected_form = form
        return selected_form['class'](self)

    def get_type_name(self):
        return settings.RESOURCE_TYPE_CONFIGS()[self.entitytypeid]['name']

    def get_primary_name(self):
        if self.entityid == '':
            return _('New Resource')
        return _('Unnamed Resource')

    def get_names(self):
        return []

    def create_resource_relationship(self, related_resource_id, notes=None, date_started=None, date_ended=None, relationship_type_id=None):
        """
        Creates a relationship between resources 
        
        """

        relationship = archesmodels.RelatedResource(
            entityid1 = self.entityid,
            entityid2 = related_resource_id,
            notes = notes,
            relationshiptype = relationship_type_id,
            datestarted = date_started,
            dateended = date_ended,
        )

        relationship.save()

    def delete_all_resource_relationships(self):
        """
        Deletes all relationships to other resources. 
        
        """     

        relationships = archesmodels.RelatedResource.objects.filter( Q(entityid2=self.entityid)|Q(entityid1=self.entityid) )

        for relationship in relationships:
            relationship.delete()

    def delete_resource_relationship(self, related_resource_id, relationship_type_id=None):
        """
        Deletes the relationships from this entity to another entity. 
        
        """        

        if relationship_type:
            relationships = archesmodels.RelatedResource.objects.filter( Q(entityid2=self.entityid)|Q(entityid1=self.entityid), Q(entityid2=related_resource_id)|Q(entityid1=related_resource_id), Q(relationshiptype=relationship_type_id))
        else:
            relationships = archesmodels.RelatedResource.objects.filter( Q(entityid2=self.entityid)|Q(entityid1=self.entityid), Q(entityid2=related_resource_id)|Q(entityid1=related_resource_id) )

        for relationship in relationships:
            relationship.delete()

    def get_related_resources(self, entitytypeid=None, relationship_type_id=None, return_entities=True):
        """
        Gets a list of entities related to this entity, optionaly filters on entitytypeid and/or relationship type. 
        Setting return_entities to False will return the relationship records 
        rather than the related entities. 
        """

        ret = []

        if self.entityid:
            if relationship_type_id:
                relationships = archesmodels.RelatedResource.objects.filter(Q(entityid2=self.entityid)|Q(entityid1=self.entityid), Q(relationshiptype=relationship_type_id))
            else:
                relationships = archesmodels.RelatedResource.objects.filter(Q(entityid2=self.entityid)|Q(entityid1=self.entityid))
      
            for relationship in relationships: 
                related_resource_id = relationship.entityid1 if relationship.entityid1 != self.entityid else relationship.entityid2
                entity_obj = archesmodels.Entities.objects.get(pk = related_resource_id)
                if (entitytypeid == None or entity_obj.entitytypeid_id == entitytypeid) and (relationship_type_id == None or relationship_type_id == relationship.relationshiptype):
                    if return_entities == True:
                        related_entity = Resource().get(related_resource_id)
                        ret.append(related_entity)
                    else:
                        ret.append(relationship)
        return ret   

    def get_alternate_names(self):
        """
        Gets the human readable name to display for entity instances

        """

        pass

    def bulk_index(self, resources=[]):
        report_documents = []
        search_documents = []
        geojson_documents = []
        terms = []

        for resource in resources:
            report_documents = report_documents + resource.prepare_documents_for_report_index()
            search_documents = search_documents + resource.prepare_documents_for_search_index()
            geojson_documents = geojson_documents + resource.prepare_documents_for_map_index(geom_entities=document['geometries'])


    def index(self):
        """
        Indexes all the nessesary documents related to resources to support the map, search, and reports

        """

        se = SearchEngineFactory().create()

        search_documents = self.prepare_documents_for_search_index()
        for document in search_documents:
            se.index_data('entity', self.entitytypeid, document, id=self.entityid)

            report_documents = self.prepare_documents_for_report_index(geom_entities=document['geometries'])
            for report_document in report_documents:
                se.index_data('resource', self.entitytypeid, report_document, id=self.entityid)

            geojson_documents = self.prepare_documents_for_map_index(geom_entities=document['geometries'])
            for geojson in geojson_documents:
                se.index_data('maplayers', self.entitytypeid, geojson, idfield='id')

        for term in self.prepare_terms_for_search_index():
           se.index_term(term['term'], term['entityid'], term['context'], term['options'])

    def prepare_documents_for_search_index(self):
        """
        Generates a list of specialized resource based documents to support resource search

        """

        document = Entity()
        document.property = self.property
        document.entitytypeid = self.entitytypeid
        document.entityid = self.entityid
        document.value = self.value
        document.label = self.label
        document.businesstablename = self.businesstablename
        document.primaryname = self.get_primary_name()
        document.child_entities = []
        document.dates = []
        document.domains = []
        document.geometries = []

        for entity in self.flatten():
            if entity.entityid != self.entityid:
                if entity.businesstablename == 'domains':
                    value = archesmodels.Values.objects.get(pk=entity.value)
                    entity_copy = entity.copy()
                    entity_copy.conceptid = value.conceptid_id
                    document.domains.append(entity_copy)
                elif entity.businesstablename == 'dates':
                    document.dates.append(entity)
                elif entity.businesstablename == 'geometries':
                    entity.value = JSONDeserializer().deserialize(fromstr(entity.value).json)
                    document.geometries.append(entity)
                else:
                    document.child_entities.append(entity)

        return [JSONSerializer().serializeToPython(document)]

    def prepare_documents_for_map_index(self, geom_entities=[]):
        """
        Generates a list of geojson documents to support the display of resources on a map

        """

        document = []
        if len(geom_entities) > 0:
            geojson_geom = {
                'type': 'GeometryCollection',
                'geometries': [geom_entity['value'] for geom_entity in geom_entities]
            }
            geom = GEOSGeometry(JSONSerializer().serialize(geojson_geom), srid=4326)
             
            document = [{
                'type': 'Feature',
                'id': self.entityid,
                'geometry':  geojson_geom,
                'properties': {
                    'entitytypeid': self.entitytypeid,
                    'primaryname': self.get_primary_name(),
                    'centroid': JSONDeserializer().deserialize(geom.centroid.json),
                    'extent': geom.extent
                }
            }]

        return document

    def prepare_terms_for_search_index(self):
        """
        Generates a list of term objects with composed of any string less then the length of settings.WORDS_PER_SEARCH_TERM  
        long and any concept associated with a resource to support term search  

        """

        terms = []

        def gather_entities(entity):
            if entity.businesstablename == '':
                pass
            elif entity.businesstablename == 'strings':
                if settings.WORDS_PER_SEARCH_TERM == None or (len(entity.value.split(' ')) < settings.WORDS_PER_SEARCH_TERM):
                    entitytype = archesmodels.EntityTypes.objects.get(pk=entity.entitytypeid)
                    terms.append({'term': entity.value, 'entityid': entity.entityid, 'context': entitytype.conceptid_id, 'options': {}})
            elif entity.businesstablename == 'domains':
                pass
            elif entity.businesstablename == 'geometries':
                pass
            elif entity.businesstablename == 'dates':
                pass
            elif entity.businesstablename == 'numbers':
                pass
            elif entity.businesstablename == 'files':
                pass

        self.traverse(gather_entities)
        return terms

    def prepare_documents_for_report_index(self, geom_entities=[]):
        """
        Generates a list of specialized resource based documents to support resource reports

        """

        geojson_geom = None
        if len(geom_entities) > 0:
            geojson_geom = {
                'type': 'GeometryCollection',
                'geometries': [geom_entity['value'] for geom_entity in geom_entities]
            }

        entity_dict = Entity()
        entity_dict.property = self.property
        entity_dict.entitytypeid = self.entitytypeid
        entity_dict.entityid = self.entityid
        entity_dict.primaryname = self.get_primary_name()
        entity_dict.geometry = geojson_geom
        
        entity_dict.graph = self.dictify(keys=['label', 'value'])
        return [JSONSerializer().serializeToPython(entity_dict)]

    def delete_index(self):
        """
        removes an entity from the search index

        """

        se = SearchEngineFactory().create()
        se.delete(index='entity', doc_type=self.entitytypeid, id=self.entityid)
        se.delete(index='resource', doc_type=self.entitytypeid, id=self.entityid)        
        se.delete(index='maplayers', doc_type=self.entitytypeid, id=self.entityid)

        def delete_indexes(entity):
            if entity.businesstablename == 'strings' or entity.businesstablename == 'domains':
                se.delete_terms(entity)

        entity = Entity().get(self.entityid)
        entity.traverse(delete_indexes)

    def prepare_search_index(self, resource_type_id, create=False):
        """
        Creates the settings and mappings in Elasticsearch to support resource search

        """

        index_settings = { 
            'settings':{
                'analysis': {
                    'analyzer': {
                        'folding': {
                            'tokenizer': 'standard',
                            'filter':  [ 'lowercase', 'asciifolding' ]
                        }
                    }
                }
            },
            'mappings': {
                resource_type_id : {
                    'properties' : {
                        'entityid' : {'type' : 'string', 'index' : 'not_analyzed'},
                        'parentid' : {'type' : 'string', 'index' : 'not_analyzed'},
                        'property' : {'type' : 'string', 'index' : 'not_analyzed'},
                        'entitytypeid' : {'type' : 'string', 'index' : 'not_analyzed'},
                        'businesstablename' : {'type' : 'string', 'index' : 'not_analyzed'},
                        'value' : {'type' : 'string', 'index' : 'not_analyzed'},
                        'label' : {'type' : 'string', 'index' : 'not_analyzed'},
                        'primaryname': {'type' : 'string', 'index' : 'not_analyzed'},
                        'child_entities' : { 
                            'type' : 'nested', 
                            'index' : 'analyzed',
                            'properties' : {
                                'entityid' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'parentid' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'property' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'entitytypeid' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'businesstablename' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'label' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'value' : {
                                    'type' : 'string',
                                    'index' : 'analyzed',
                                    'fields' : {
                                        'raw' : { 'type' : 'string', 'index' : 'not_analyzed'},
                                        'folded': { 'type': 'string', 'analyzer': 'folding'}
                                    }
                                }
                            }
                        },
                        'domains' : { 
                            'type' : 'nested', 
                            'index' : 'analyzed',
                            'properties' : {
                                'entityid' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'parentid' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'property' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'entitytypeid' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'businesstablename' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'label' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'value' : {
                                    'type' : 'string',
                                    'index' : 'analyzed',
                                    'fields' : {
                                        'raw' : { 'type' : 'string', 'index' : 'not_analyzed'}
                                    }
                                },
                                'conceptid' : {'type' : 'string', 'index' : 'not_analyzed'},
                            }
                        },
                        'geometries' : { 
                            'type' : 'nested', 
                            'index' : 'analyzed',
                            'properties' : {
                                'entityid' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'parentid' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'property' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'entitytypeid' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'businesstablename' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'label' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'value' : {
                                    "type": "geo_shape"
                                }
                            }
                        },
                        'dates' : { 
                            'type' : 'nested', 
                            'index' : 'analyzed',
                            'properties' : {
                                'entityid' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'parentid' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'property' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'entitytypeid' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'businesstablename' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'label' : {'type' : 'string', 'index' : 'not_analyzed'},
                                'value' : {
                                    "type" : "date"
                                }
                            }
                        }
                    }
                }
            }
        }

        if create:
            se = SearchEngineFactory().create()
            try:
                se.create_index(index='entity', body=index_settings)
            except:
                index_settings = index_settings['mappings']
                se.create_mapping(index='entity', doc_type=resource_type_id, body=index_settings)

        return index_settings

    def prepare_term_index(self, create=False):
        """
        Creates the settings and mappings in Elasticsearch to support term search

        """

        index_settings = {
            'settings':{
                'analysis': {
                    'analyzer': {
                        'folding': {
                            'tokenizer': 'standard',
                            'filter':  [ 'lowercase', 'asciifolding' ]
                        }
                    }
                }
            },
            'mappings':{
                'value':{
                    'properties': {
                        'ids':{'type': 'string', 'index' : 'not_analyzed'},
                        'context':{'type': 'string', 'index' : 'not_analyzed'},
                        'term': { 
                            'type': 'string',
                            'analyzer': 'standard',
                            'fields': {
                                'folded': { 
                                    'type': 'string',
                                    'analyzer': 'folding'
                                }
                            }
                        }
                    }            
                }            
            }
        }

        if create:
            se = SearchEngineFactory().create()
            se.create_index(index='term', body=index_settings, ignore=400)

        return index_settings

    def prepare_resource_relations_index(self, create=False):
        """
        Creates the settings and mappings in Elasticsearch to support related resources

        """

        index_settings = { 
            'mappings':{
                'all': {
                    'properties': {
                        'resourcexid': {'type': 'long'},
                        'notes': { 'type': 'string'},
                        'relationshiptype': {'type': 'string', 'index' : 'not_analyzed'},
                        'entityid2': {'type': 'string', 'index' : 'not_analyzed'},
                        'entityid1': {'type': 'string', 'index' : 'not_analyzed'}
                    }  
                }
            }
        }    

        if create:
            se = SearchEngineFactory().create()
            se.create_index(index='resource_relations', body=index_settings, ignore=400)

        return index_settings
           
    @staticmethod
    def get_report(resourceid):
        return {
            'id': None,
            'data': None
        }



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

from datetime import date
from django.conf import settings
from django.template import RequestContext
from django.shortcuts import render_to_response
from django.core.paginator import Paginator
from django.utils.importlib import import_module
from django.contrib.gis.geos import GEOSGeometry
from arches.app.models.concept import Concept
from arches.app.utils.JSONResponse import JSONResponse
from arches.app.utils.betterJSONSerializer import JSONSerializer, JSONDeserializer
from arches.app.views.concept import get_preflabel_from_conceptid
from arches.app.search.search_engine_factory import SearchEngineFactory
from arches.app.search.elasticsearch_dsl_builder import Bool, Match, Query, Nested, Terms, GeoShape, Range
geocoder = import_module(settings.GEOCODING_PROVIDER)

def home_page(request):
    lang = request.GET.get('lang', 'en-us')

    return render_to_response('search.htm', {
            'main_script': 'search',
            'active_page': 'Search'
        }, 
        context_instance=RequestContext(request))

def search_terms(request):
    lang = request.GET.get('lang', 'en-us')
    
    query = build_search_terms_dsl(request)
    results = query.search(index='term', doc_type='value')

    for result in results['hits']['hits']:
        prefLabel = get_preflabel_from_conceptid(result['_source']['context'], lang)
        result['_source']['context'] = prefLabel['value']

    return JSONResponse(results)

def build_search_terms_dsl(request):
    se = SearchEngineFactory().create()
    searchString = request.GET.get('q', '')
    query = Query(se, start=0, limit=settings.SEARCH_DROPDOWN_LENGTH)
    boolquery = Bool()
    boolquery.should(Match(field='term', query=searchString.lower(), type='phrase_prefix', fuzziness='AUTO'))
    boolquery.should(Match(field='term.folded', query=searchString.lower(), type='phrase_prefix', fuzziness='AUTO'))
    boolquery.should(Match(field='term.folded', query=searchString.lower(), fuzziness='AUTO'))
    query.add_query(boolquery)

    return query

def search_results(request, as_text=False):
    dsl = build_search_results_dsl(request)
    results = dsl.search(index='entity', doc_type='') 
    total = results['hits']['total']
    page = 1 if request.GET.get('page') == '' else int(request.GET.get('page', 1))
    all_entity_ids = ['_all']
    if request.GET.get('include_ids', 'false') == 'false':
        all_entity_ids = ['_none']
    elif request.GET.get('no_filters', '') == '':
        full_dsl = build_search_results_dsl(request)
        full_results = full_dsl.search(index='entity', doc_type='', start=0, limit=1000000, fields=[])
        all_entity_ids = [hit['_id'] for hit in full_results['hits']['hits']]

    return _get_pagination(results, total, page, settings.SEARCH_ITEMS_PER_PAGE, all_entity_ids)

def build_search_results_dsl(request):
    term_filter = request.GET.get('termFilter', '')
    spatial_filter = JSONDeserializer().deserialize(request.GET.get('spatialFilter', None)) 
    export = request.GET.get('export', None)
    page = 1 if request.GET.get('page') == '' else int(request.GET.get('page', 1))
    temporal_filter = JSONDeserializer().deserialize(request.GET.get('temporalFilter', None))

    se = SearchEngineFactory().create()

    if export != None:
        limit = settings.SEARCH_EXPORT_ITEMS_PER_PAGE  
    else:
        limit = settings.SEARCH_ITEMS_PER_PAGE
    
    query = Query(se, start=limit*int(page-1), limit=limit)
    boolquery = Bool()
    boolfilter = Bool()
    
    if term_filter != '':
        for term in JSONDeserializer().deserialize(term_filter):
            if term['type'] == 'term':
                boolfilter_folded = Bool()
                boolfilter_folded.should(Match(field='child_entities.value', query=term['value'], type='phrase'))
                boolfilter_folded.should(Match(field='child_entities.value.folded', query=term['value'], type='phrase'))
                nested = Nested(path='child_entities', query=boolfilter_folded)
                if term['inverted']:
                    boolfilter.must_not(nested)
                else:    
                    boolfilter.must(nested)
            elif term['type'] == 'concept':
                concept_ids = _get_child_concepts(term['value'])
                terms = Terms(field='domains.conceptid', terms=concept_ids)
                nested = Nested(path='domains', query=terms)
                if term['inverted']:
                    boolfilter.must_not(nested)
                else:
                    boolfilter.must(nested)
            elif term['type'] == 'string':
                boolfilter_folded = Bool()
                boolfilter_folded.should(Match(field='child_entities.value', query=term['value'], type='phrase_prefix'))
                boolfilter_folded.should(Match(field='child_entities.value.folded', query=term['value'], type='phrase_prefix'))
                #phrase = Match(field='child_entities.value', query=term['value'], type='phrase_prefix')
                nested = Nested(path='child_entities', query=boolfilter_folded)
                if term['inverted']:
                    boolquery.must_not(nested)
                else:    
                    boolquery.must(nested)

    if 'geometry' in spatial_filter and 'type' in spatial_filter['geometry'] and spatial_filter['geometry']['type'] != '':
        geojson = spatial_filter['geometry']
        if geojson['type'] == 'bbox':
            coordinates = [[geojson['coordinates'][0],geojson['coordinates'][3]], [geojson['coordinates'][2],geojson['coordinates'][1]]]
            geoshape = GeoShape(field='geometries.value', type='envelope', coordinates=coordinates )
            nested = Nested(path='geometries', query=geoshape)
        else:
            buffer = spatial_filter['buffer']
            geojson = JSONDeserializer().deserialize(_buffer(geojson,buffer['width'],buffer['unit']).json)
            geoshape = GeoShape(field='geometries.value', type=geojson['type'], coordinates=geojson['coordinates'] )
            nested = Nested(path='geometries', query=geoshape)

        if 'inverted' not in spatial_filter:
            spatial_filter['inverted'] = False

        if spatial_filter['inverted']:
            boolfilter.must_not(nested)
        else:
            boolfilter.must(nested)

    if 'year_min_max' in temporal_filter and len(temporal_filter['year_min_max']) == 2:
        start_date = date(temporal_filter['year_min_max'][0], 1, 1)
        end_date = date(temporal_filter['year_min_max'][1], 12, 31)
        if start_date:
            start_date = start_date.strftime('%Y-%m-%d')
        if end_date:
            end_date = end_date.strftime('%Y-%m-%d')
        range = Range(field='dates.value', gte=start_date, lte=end_date)
        nested = Nested(path='dates', query=range)
        
        if 'inverted' not in temporal_filter:
            temporal_filter['inverted'] = False

        if temporal_filter['inverted']:
            boolfilter.must_not(nested)
        else:
            boolfilter.must(nested)
        
    if not boolquery.empty:
        query.add_query(boolquery)

    if not boolfilter.empty:
        query.add_filter(boolfilter)

    return query

def buffer(request):
    spatial_filter = JSONDeserializer().deserialize(request.GET.get('filter', {'geometry':{'type':'','coordinates':[]},'buffer':{'width':'0','unit':'ft'}})) 

    if spatial_filter['geometry']['coordinates'] != '' and spatial_filter['geometry']['type'] != '':
        return JSONResponse(_buffer(spatial_filter['geometry'],spatial_filter['buffer']['width'],spatial_filter['buffer']['unit']), geom_format='json')

    return JSONResponse()

def _buffer(geojson, width=0, unit='ft'):
    geojson = JSONSerializer().serialize(geojson)
    
    try:
        width = float(width)
    except:
        width = 0

    if width > 0:
        geom = GEOSGeometry(geojson, srid=4326)
        geom.transform(3857)

        if unit == 'ft':
            width = width/3.28084

        buffered_geom = geom.buffer(width)
        buffered_geom.transform(4326)
        return buffered_geom
    else:
        return GEOSGeometry(geojson)

def _get_child_concepts(conceptid):
    ret = set([conceptid])
    for row in Concept().get_child_concepts(conceptid, 'narrower', ['prefLabel'], 'prefLabel'):
        ret.add(row[0])
        ret.add(row[1])
    return list(ret)

def _get_pagination(results, total_count, page, count_per_page, all_ids):
    paginator = Paginator(range(total_count), count_per_page)
    pages = [page]
    if paginator.num_pages > 1:
        before = paginator.page_range[0:page-1]
        after = paginator.page_range[page:paginator.num_pages]
        default_ct = 3
        ct_before = default_ct if len(after) > default_ct else default_ct*2-len(after)
        ct_after = default_ct if len(before) > default_ct else default_ct*2-len(before)
        if len(before) > ct_before:
            before = [1,None]+before[-1*(ct_before-1):]
        if len(after) > ct_after:
            after = after[0:ct_after-1]+[None,paginator.num_pages]
        pages = before+pages+after
    return render_to_response('pagination.htm', {'pages': pages, 'page_obj': paginator.page(page), 'results': JSONSerializer().serialize(results), 'all_ids': JSONSerializer().serialize(all_ids)})

def geocode(request):
    search_string = request.GET.get('q', '')    
    return JSONResponse({ 'results': geocoder.find_candidates(search_string) })



import xml.etree.ElementTree as ET
from django.http import StreamingHttpResponse
from django.http import HttpResponse
import json
import os
import zipfile
from datetime import datetime
import glob
import csv
import shapefile
import arches.app.utils.geos_to_pyshp as geos_to_pyshp
import codecs

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from django.contrib.gis.geos import GEOSGeometry
from django.contrib.gis.geos import GeometryCollection

def export_results(request):
    '''
    Takes a search request with an export format parameter and returns the respective format.

    '''

    #search_results = search_resources(request, district_number) 
    dsl = build_search_results_dsl(request)
    search_results = dsl.search(index='entity', doc_type='') 
    response = None
    format = request.GET.get('export', 'csv')

    if len(search_results['hits']['hits']) > 0:

        if format == 'geojson':
            geo_json = format_search_as_geojson(search_results, geojson=True)
            return JSONResponse(geo_json, indent=4)  

        elif format == 'shp':
            shp_prepped = format_search_as_geojson(search_results)
            shapefiles = create_shapefiles(shp_prepped)
            return shapefiles

        elif format == 'kml':
            kml_prepped = format_search_as_geojson(search_results)
            kml = create_kml(kml_prepped)
            response = HttpResponse(kml, content_type='text/xml')
            response['Content-Disposition'] = 'attachment; filename="search_result_export.kml"'

        elif format == 'csv':
            csv_prepped = format_search_as_geojson(search_results)
            csvs = create_csvs(csv_prepped)
            return csvs

            # Leave this here for now - in case we want to return a single csv file rather than a zip file

            # feature_collection = format_search_as_geojson(search_results)
            # prepare_caltrans_csv(feature_collection)
            # fieldnames = feature_collection['features'][0]['properties'].keys()
            # csvfile = StringIO()
            # csvwriter = csv.DictWriter(csvfile, delimiter='|', fieldnames=fieldnames)

            # def read_and_flush():
            #     csvfile.seek(0)
            #     data = csvfile.read()
            #     csvfile.seek(0)
            #     csvfile.truncate()
            #     return data

            # def stream():
            #     csvwriter.writeheader()
            #     for feature in feature_collection['features']:
            #         csvwriter.writerow(feature['properties'])
            #     data = read_and_flush()
            #     yield data

            # response = StreamingHttpResponse(stream(), content_type='text/csv')
            # response['Content-Disposition'] = 'attachment; filename="search_result_export.csv"'
            # return response

    else:
        response = HttpResponse('no search results available', content_type='text')
        response['Content-Disposition'] = 'attachment; filename="search_result_export.txt"'

    return response

def format_search_as_geojson(search_results, geojson=False):
    """
    Takes a 2d list of search results, groups them in to dictionaries by their resourceid, 
    and returns a geojson-like dictionary. If geojson is set to True, the returned object 
    is a valid geojson dictionary. 

    """

    resource_mappings = read_export_configs()
    features = []
    resources = search_results['hits']['hits']

    for resource in resources:
        if resource['_type'] in resource_mappings:
            source = resource['_source']
            field_mapping = resource_mappings[source['entitytypeid']]['fieldmap']
            
            properties = {}
            for entitytypeid, field_name in field_mapping.iteritems():
                print field_name
                properties[field_name] = None #A shapefile needs every possible field whether a resource has a matching related entity or not. ...So we add each field with None for every record.

            properties['type'] = source['entitytypeid']
            properties['id'] = resource['_id']

            #raise Exception()

            for entity in source['dates']:
                if entity['entitytypeid'] in field_mapping:
                    field_name = field_mapping[entity['entitytypeid']]
                    properties[field_name] = entity['value']

            geoms = []

            if len(source['geometries']) > 0:
                for g in source['geometries']:
                    geom = GEOSGeometry(JSONSerializer().serialize(g['value'], ensure_ascii=False))
                    geoms.append(geom)

            geometry = GeometryCollection(geoms)

            if geojson == True:
                geometry = json.loads(geometry.geojson)

            feature = {
                'type':'Feature',
                'geometry': geometry,
                'properties': properties
                }

            features.append(feature)

    feature_collection = {'type':'FeatureCollection', 'features': features}

    return feature_collection

def create_kml(feature_collection):
    """
    Takes a list of geojson dictionaries - with geometry still as a geos object and converts them to an ElementTree object with KML tags
    and attributes. Returns a kml version of the object.

    """

    root = ET.Element("kml", attrib={'xmlns':'http://earth.google.com/kml/2.0'})
    folder = ET.SubElement(root, "Folder")
    name = ET.SubElement(folder, "name")
    name.text = 'search-result-export'
    description = ET.SubElement(folder, 'description')
    description.text = 'Search Export Results'
    style = ET.SubElement(folder, 'Style', attrib={'id':'resource_style'})
    linestyle = ET.SubElement(style, 'LineStyle')
    width = ET.SubElement(linestyle, 'width')
    width.text = '2.0'
    line_color = ET.SubElement(linestyle, 'color')
    line_color.text = '7d0000ff'
    polystyle = ET.SubElement(style, 'PolyStyle')
    color = ET.SubElement(polystyle, 'color')
    color.text = '7d0000ff'
    colormode = ET.SubElement(polystyle, 'colorMode')
    colormode.text = 'normal'
    
    for feature in feature_collection['features']:
        placemark = ET.SubElement(folder,'Placemark')
        placemark_name = ET.SubElement(placemark, 'name')
        placemark_name.text = feature['properties']['type']
        placemark_description = ET.SubElement(placemark, 'description')
        visibility = ET.SubElement(placemark, 'visibility')
        visibility.text = '1'
        style_url = ET.SubElement(placemark, 'styleUrl')
        style_url.text = '#resource_style'

        extended_data = ET.SubElement(placemark, 'ExtendedData')
        for k, v in feature['properties'].iteritems():
            data = ET.SubElement(extended_data, 'Data', attrib={'name':k})
            value = ET.SubElement(data, 'value')
            value.text = v
        geometry_element = ET.fromstring(feature['geometry'].kml)
        placemark.append(geometry_element)

    return ET.tostring(root, encoding='UTF-8', method='xml')


def create_shapefiles(feature_collection):
    '''
    Takes a geojson-like (geojson-like because it has a geos geometry rather than a geojson geometry, allowing us to modify the 
    geometry to be a centroid or hull if necessary.) feature collection, groups the data by resource type and creates a shapefile
    for each resource. Returns a .zip file with each of the shapefiles. An arches export configuration file is needed to map shapefile
    fields to resorce entitytypeids and specify the shapefile column datatypes (fiona schema). 

    '''

    resource_export_groups = read_export_configs()

    for feature in feature_collection['features']:
        feature['geometry'] = feature['geometry'].centroid
        resource_export_groups[feature['properties']['type']]['records'].append(feature)      

    shapefiles_for_export = []
    for feature_class, data in resource_export_groups.iteritems():
        geom_type = data['schema']['geometry']

        if len(data['records']) > 0:
            if geom_type == 'Point':
                writer = shapefile.Writer(shapeType=shapefile.MULTIPOINT)
            elif geom_type == 'LineString':
                writer = shapefile.Writer(shapeType=shapefile.POLYLINE)
            elif geom_type == 'Polygon':
                writer = shapefile.Writer(shapeType=shapefile.POLYGON)

            geos_datatypes_to_pyshp_types = {'str':'C', 'datetime':'D', 'float':'F'}

            for k, v in data["schema"]["properties"].iteritems():
                writer.field(codecs.encode(k), geos_datatypes_to_pyshp_types[v])

            for r in data['records']:
                shp_geom = geos_to_pyshp.convert_geom(r['geometry'])
                if geom_type in ['Point','LineString']:
                    writer.line(parts=shp_geom)
                elif geom_type == 'Polygon':
                    writer.poly(parts=shp_geom)
                writer.record(**r['properties'])

            export_file_name = os.path.join(settings.PACKAGE_ROOT, 'tmp', data['export_file_name'], data['export_file_name'])
            writer.save(export_file_name)
            shapefiles_for_export.append(export_file_name)

    resp = zip_response(shapefiles_for_export, zip_file_name='search_result_shapefile_export.zip')
    return resp


def create_csvs(feature_collection):
    '''
    Groups data by resource type as defined in the arches_export_file and creates a csv file for each resource type. 
    Returns a .zip file with each of the csv files. An arches export configuration file is needed to map shapefile
    fields to resorce entitytypeids. 

    '''

    resource_export_groups = read_export_configs()

    for feature in feature_collection['features']:
        resource_export_groups[feature['properties']['type']]['records'].append(feature)   

    csvs_for_export = []
    for feature_class, data in resource_export_groups.iteritems():
        if len(data['records']) > 0:
            if not os.path.exists(os.path.join(settings.PACKAGE_ROOT, 'tmp')):
                os.makedirs(os.path.join(settings.PACKAGE_ROOT, 'tmp'))
            csv_name = os.path.join(settings.PACKAGE_ROOT, 'tmp', data['export_file_name'])
            csvs_for_export.append(csv_name)
            with open(csv_name + '.csv', 'w') as dest:
                fieldnames = data['schema']['properties'].keys()
                csvwriter = csv.DictWriter(dest, delimiter='|', fieldnames=fieldnames)
                csvwriter.writeheader()
                for record in data['records']:
                    csvwriter.writerow(record['properties'])

    response = zip_response(csvs_for_export, zip_file_name='search_result_csv_export.zip', file_type='.csv')
    return response

    #Do not delete for now - This was written to support building flat records from search results
    # for project in projects:
    #     project_record = {'id': project['properties']['projectid'], 'project':project, 'documents': []}
    #     for document in documents:
    #         if document['properties']['projectid'] == project_record['id']:
    #             document_record = {'id': document['properties']['documentid'], 'document':document, 'mitigations': []}
    #             for mitigation in mitigations:
    #                 if mitigation['properties']['documentid'] == document_record['id']:
    #                     document_record['mitigations'].append(mitigation)
    #             project_record['documents'].append(document_record)
    #     project_records.append(project_record)
    
def read_export_configs():
    '''
    Reads the export configuration file and adds an array for records to store property data

    '''

    resource_export_groups = json.load(open(settings.EXPORT_CONFIG, 'r'))
    for key, val in resource_export_groups.iteritems():
        resource_export_groups[key]['records'] = []

    return resource_export_groups


def zip_response(files_for_export, zip_file_name=None, file_type=None):
    '''
    Given a list of export file names, zips up all the files with those names and returns and http response.

    '''

    d = datetime.now()
    timestamp = '{0}{1}{2}_{3}{4}'.format(d.year, d.month, d.day, d.hour, d.minute) #Should we append the timestamp to the exported filename?

    buffer = StringIO()
    zip = zipfile.ZipFile(buffer, 'w', zipfile.ZIP_DEFLATED)

    for filename in files_for_export:
        if file_type:
            for f in glob.glob(filename + '*' + file_type):
                zip.write(f, os.path.basename(f))
        else:
            for f in glob.glob(filename + '*'):
                zip.write(f, os.path.basename(f))

    zip.close()
    buffer.flush()
    zip_stream = buffer.getvalue()
    buffer.close()

    response = HttpResponse()
    response['Content-Disposition'] = 'attachment; filename=' + zip_file_name
    response['Content-length'] = str(len(zip_stream))
    response['Content-Type'] = 'application/zip'
    response.write(zip_stream)
    return response



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

import urllib
import uuid
import logging
from django.conf import settings
from datetime import datetime
from elasticsearch import Elasticsearch, helpers


class SearchEngine(object):

    def __init__(self):
        self.es = Elasticsearch(hosts=settings.ELASTICSEARCH_HOSTS, **settings.ELASTICSEARCH_CONNECTION_OPTIONS)
        self.logger = logging.getLogger(__name__)

    def delete(self, **kwargs):
        """
        Deletes a document from the index
        Pass an index, doc_type, and id to delete a specific document
        Pass a body with a query dsl to delete by query

        """

        body = kwargs.get('body', None)
        if body != None:
            try:
                return self.es.delete_by_query(ignore=[404], **kwargs)
            except Exception as detail:
                self.logger.warning('%s: WARNING: failed to delete document by query: %s \nException detail: %s\n' % (datetime.now(), body, detail))
                raise detail   
        else:
            try:
                return self.es.delete(ignore=[404], **kwargs)
            except Exception as detail:
                self.logger.warning('%s: WARNING: failed to delete document: %s \nException detail: %s\n' % (datetime.now(), body, detail))
                raise detail   

    def delete_index(self, **kwargs):
        """
        Deletes an entire index

        """

        index = kwargs.get('index', '').strip()
        print 'deleting index : %s' % index
        return self.es.indices.delete(index=index, ignore=[400, 404])

    def search(self, **kwargs):
        """
        Search for an item in the index.
        Pass an index, doc_type, and id to get a specific document
        Pass a body with a query dsl to perform a search

        """

        body = kwargs.get('body', None)
        index = kwargs.get('index', None)
        id = kwargs.get('id', None)

        if index is None:
            raise NotImplementedError("You must specify an 'index' in your call to search")

        if id:
            if isinstance(id, list):
                kwargs.setdefault('body', {'ids': kwargs.pop('id')})
                return self.es.mget(**kwargs)
            else:
                return self.es.get(**kwargs)
        
        ret = None
        try: 
            ret = self.es.search(**kwargs)
        except Exception as detail:
            self.logger.warning('%s: WARNING: search failed for query: %s \nException detail: %s\n' % (datetime.now(), body, detail))
            pass   

        return ret

    def index_term(self, term, id, context='', options={}):
        """
        If the term is already indexed, then simply increment the count and add the entity id of the term to the existing index.
        If the term isn't indexed then add the index.

        """

        if term.strip(' \t\n\r') != '':
            already_indexed = False
            count = 1
            ids = [id]
            
            try:
                #_id = unicode(term, errors='ignore').decode('utf-8').encode('ascii')
                _id = uuid.uuid3(uuid.NAMESPACE_DNS, '%s%s' % (hash(term), hash(context)))
                result = self.es.get(index='term', doc_type='value', id=_id, ignore=404)

                #print 'result: %s' % result
                if result['found'] == True:
                    ids = result['_source']['ids']
                    if id not in ids:
                        ids.append(id)
                else:
                    ids = [id]

                self.index_data('term', 'value', {'term': term, 'context': context, 'options': options, 'count': len(ids), 'ids': ids}, id=_id)

            except Exception as detail:
                self.logger.warning('%s: WARNING: search failed to index term: %s \nException detail: %s\n' % (datetime.now(), term, detail))
                raise detail   
                  
    def delete_terms(self, entities):
        """
        If the term is referenced more then once simply decrement the 
        count and remove the entity id of the deleted term from the from the existing index.

        If the term is only referenced once then delete the index  

        """

        if not isinstance(entities, list):
            entities = [entities]

        for entity in entities:
            result = self.es.search(index='term', doc_type='value', body={
                "query": {
                    "filtered": {
                        "filter":{
                            "terms": {
                                "ids": [entity.entityid]
                            }
                        }, 
                        "query": {
                            "match_all": {}
                        }
                    }
                }, 
                "from": 0, 
                "size": 10
            }, ignore=404)

            if 'hits' in result:
                for document in result['hits']['hits']:
                    document['_source']['ids'].remove(entity.entityid)
                    count = len(document['_source']['ids'])
                    if count > 0:
                        document['_source']['count'] = count
                        self.index_data('term', 'value', document['_source'], id=document['_id'])
                    else:
                        self.delete(index='term', doc_type='value', id=document['_id'])

    def create_mapping(self, index, doc_type, fieldname='', fieldtype='string', fieldindex='analyzed', body=None):
        """
        Creates an Elasticsearch body for a single field given an index name and type name

        """

        if not body:
            if fieldtype == 'geo_shape':
                body =  { 
                    doc_type : {
                        'properties' : {
                            fieldname : { 'type' : 'geo_shape', 'tree' : 'geohash', 'precision': '1m' }
                        }
                    }
                } 
            else:           
                body =  { 
                    doc_type : {
                        'properties' : {
                            fieldname : { 'type' : fieldtype, 'index' : fieldindex }
                        }
                    }
                }

        self.create_index(index=index, ignore=400)
        self.es.indices.put_mapping(index=index, doc_type=doc_type, body=body)

    def create_index(self, **kwargs):
        self.es.indices.create(**kwargs)

    def index_data(self, index=None, doc_type=None, body=None, idfield=None, id=None, **kwargs):
        """
        Indexes a document or list of documents into Elasticsearch

        If "id" is supplied then will use that as the id of the document

        If "idfield" is supplied then will try to find that property in the 
            document itself and use the value found for the id of the document

        """

        if not isinstance(body, list):
            body = [body]

        for document in body:
            if idfield is not None:
                if isinstance(document, dict):
                    id = document[idfield]
                else:
                    id = getattr(document,idfield)

            try:
                self.es.index(index=index, doc_type=doc_type, body=document, id=id, **kwargs)
            except Exception as detail:
                self.logger.warning('%s: WARNING: failed to index document: %s \nException detail: %s\n' % (datetime.now(), document, detail))
                raise detail


    def bulk_index(self, data):
        return helpers.bulk(self.es, data, chunk_size=500, raise_on_error=True)

    def create_bulk_item(self, index, type, id, data):
        if not(self.isempty_or_none(index) or self.isempty_or_none(type) or self.isempty_or_none(id)):
            return[
                { "index" : { "_index" : index, "_type" : type, "_id" : id } },
                data
            ]
        else:
            return false


'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

"""This module contains commands for building Arches."""

from optparse import make_option
from django.core.management.base import BaseCommand, CommandError
from django.conf import settings
from django.utils.importlib import import_module
import os, sys, subprocess
from arches.setup import get_elasticsearch_download_url, download_elasticsearch, unzip_file
from arches.db.install import truncate_db, install_db
from arches.app.utils.data_management.resources.importer import ResourceLoader
import arches.app.utils.data_management.resources.remover as resource_remover
from arches.management.commands import utils
from arches.app.search.search_engine_factory import SearchEngineFactory

class Command(BaseCommand):
    """
    Commands for managing the loading and running of packages in Arches

    """
    
    option_list = BaseCommand.option_list + (
        make_option('-o', '--operation', action='store', dest='operation', default='setup',
            type='choice', choices=['setup', 'install', 'setup_db', 'start_elasticsearch', 'build_permissions', 'livereload', 'load_resources', 'remove_resources'],
            help='Operation Type; ' +
            '\'setup\'=Sets up Elasticsearch and core database schema and code' + 
            '\'setup_db\'=Truncate the entire arches based db and re-installs the base schema' + 
            '\'install\'=Runs the setup file defined in your package root' + 
            '\'start_elasticsearch\'=Runs the setup file defined in your package root' + 
            '\'build_permissions\'=generates "add,update,read,delete" permissions for each entity mapping'+
            '\'livereload\'=Starts livereload for this package on port 35729'),
        make_option('-s', '--source', action='store', dest='source', default='',
            help='Directory containing a .arches or .shp file containing resource records'),
        make_option('-f', '--format', action='store', dest='format', default='arches',
            help='Format: shp or arches'),
        make_option('-l', '--load_id', action='store', dest='load_id',
            help='Text string identifying the resources in the data load you want to delete.'),
    )

    def handle(self, *args, **options):
        print 'operation: '+ options['operation']
        package_name = settings.PACKAGE_NAME
        print 'package: '+ package_name
        
        if options['operation'] == 'setup':
            self.setup(package_name)

        if options['operation'] == 'install':
            self.install(package_name)

        if options['operation'] == 'setup_db':
            self.setup_db(package_name)

        if options['operation'] == 'start_elasticsearch':
            self.start_elasticsearch(package_name)

        if options['operation'] == 'livereload':
            self.start_livereload()

        if options['operation'] == 'build_permissions':
            self.build_permissions()

        if options['operation'] == 'load_resources':
            self.load_resources(package_name, options['source'])
            
        if options['operation'] == 'remove_resources':     
            self.remove_resources(options['load_id'])


    def setup(self, package_name):
        """
        Installs Elasticsearch into the package directory and 
        installs the database into postgres as "arches_<package_name>"

        """
        self.setup_elasticsearch(package_name, port=settings.ELASTICSEARCH_HTTP_PORT)  
        self.setup_db(package_name)
        self.generate_procfile(package_name)

    def install(self, package_name):
        """
        Runs the setup.py file found in the package root

        """

        module = import_module('%s.setup' % package_name)
        install = getattr(module, 'install')
        install() 

    def setup_elasticsearch(self, package_name, port=9200):
        """
        Installs Elasticsearch into the package directory and
        adds default settings for running in a test environment

        Change these settings in production

        """

        install_location = self.get_elasticsearch_install_location(package_name)
        install_root = os.path.abspath(os.path.join(install_location, '..'))
        url = get_elasticsearch_download_url(os.path.join(settings.ROOT_DIR, 'install'))
        file_name = url.split('/')[-1]

        try:
            unzip_file(os.path.join(settings.ROOT_DIR, 'install', file_name), install_root)
        except:
            download_elasticsearch(os.path.join(settings.ROOT_DIR, 'install'))

        es_config_directory = os.path.join(install_location, 'config')
        try:
            os.rename(os.path.join(es_config_directory, 'elasticsearch.yml'), os.path.join(es_config_directory, 'elasticsearch.yml.orig'))
        except: pass

        with open(os.path.join(es_config_directory, 'elasticsearch.yml'), 'w') as f:
            f.write('# ----------------- FOR TESTING ONLY -----------------')
            f.write('\n# - THESE SETTINGS SHOULD BE REVIEWED FOR PRODUCTION -')
            f.write('\nnode.max_local_storage_nodes: 1')
            f.write('\nindex.number_of_shards: 1')
            f.write('\nindex.number_of_replicas: 0')
            f.write('\nhttp.port: %s' % port)
            f.write('\ndiscovery.zen.ping.multicast.enabled: false')
            f.write('\ndiscovery.zen.ping.unicast.hosts: ["localhost"]')
            f.write('\ncluster.routing.allocation.disk.threshold_enabled: false')

        # install plugin
        if sys.platform == 'win32':
            os.system("call %s --install mobz/elasticsearch-head" % (os.path.join(install_location, 'bin', 'plugin.bat')))
        else:
            os.chdir(os.path.join(install_location, 'bin'))
            os.system("chmod u+x plugin")
            os.system("./plugin -install mobz/elasticsearch-head")
            os.system("chmod u+x elasticsearch")

    def start_elasticsearch(self, package_name):
        """
        Starts the Elasticsearch process (blocking)
        WARNING: this will block all subsequent python calls

        """

        es_start = os.path.join(self.get_elasticsearch_install_location(package_name), 'bin')
        
        # use this instead to start in a non-blocking way
        if sys.platform == 'win32':
            import time
            p = subprocess.Popen(['service.bat', 'install'], cwd=es_start, shell=True)  
            time.sleep(10)
            p = subprocess.Popen(['service.bat', 'start'], cwd=es_start, shell=True) 
        else:
            p = subprocess.Popen(es_start + '/elasticsearch', cwd=es_start, shell=False)  
        return p
        #os.system('honcho start')

    def setup_db(self, package_name):
        """
        Drops and re-installs the database found at "arches_<package_name>"
        WARNING: This will destroy data

        """

        db_settings = settings.DATABASES['default']
        truncate_path = os.path.join(settings.ROOT_DIR, 'db', 'install', 'truncate_db.sql')
        install_path = os.path.join(settings.ROOT_DIR, 'db', 'install', 'install_db.sql')  
        db_settings['truncate_path'] = truncate_path
        db_settings['install_path'] = install_path   
        
        truncate_db.create_sqlfile(db_settings, truncate_path)
        install_db.create_sqlfile(db_settings, install_path)
        
        os.system('psql -h %(HOST)s -p %(PORT)s -U %(USER)s -d postgres -f "%(truncate_path)s"' % db_settings)
        os.system('psql -h %(HOST)s -p %(PORT)s -U %(USER)s -d %(NAME)s -f "%(install_path)s"' % db_settings)

    def generate_procfile(self, package_name):
        """
        Generate a procfile for use with Honcho (https://honcho.readthedocs.org/en/latest/)

        """

        python_exe = os.path.abspath(sys.executable)

        contents = []
        contents.append('\nelasticsearch: %s' % os.path.join(self.get_elasticsearch_install_location(package_name), 'bin', 'elasticsearch'))
        contents.append('django: %s manage.py runserver' % (python_exe))
        contents.append('livereload: %s manage.py packages --operation livereload' % (python_exe))

        package_root = settings.PACKAGE_ROOT
        if hasattr(settings, 'SUBPACKAGE_ROOT'):
            package_root = settings.SUBPACKAGE_ROOT

        utils.write_to_file(os.path.join(package_root, '..', 'Procfile'), '\n'.join(contents))

    def get_elasticsearch_install_location(self, package_name):
        """
        Get the path to the Elasticsearch install

        """

        url = get_elasticsearch_download_url(os.path.join(settings.ROOT_DIR, 'install'))
        file_name = url.split('/')[-1]
        file_name_wo_extention = file_name[:-4]
        package_root = settings.PACKAGE_ROOT
        return os.path.join(package_root, 'elasticsearch', file_name_wo_extention)

    def build_permissions(self):
        """
        Creates permissions based on all the installed resource types

        """

        from arches.app.models import models
        from django.contrib.auth.models import Permission, ContentType

        resourcetypes = {}
        mappings = models.Mappings.objects.all()
        mapping_steps = models.MappingSteps.objects.all()
        rules = models.Rules.objects.all()
        for mapping in mappings:
            #print '%s -- %s' % (mapping.entitytypeidfrom_id, mapping.entitytypeidto_id)
            if mapping.entitytypeidfrom_id not in resourcetypes:
                resourcetypes[mapping.entitytypeidfrom_id] = set([mapping.entitytypeidfrom_id])
            for step in mapping_steps.filter(pk=mapping.pk):
                resourcetypes[mapping.entitytypeidfrom_id].add(step.ruleid.entitytyperange_id)

        for resourcetype in resourcetypes:
            for entitytype in resourcetypes[resourcetype]:
                content_type = ContentType.objects.get_or_create(name='Arches', app_label=resourcetype, model=entitytype)
                Permission.objects.create(codename='add_%s' % entitytype, name='%s - add' % entitytype , content_type=content_type[0])
                Permission.objects.create(codename='update_%s' % entitytype, name='%s - update' % entitytype , content_type=content_type[0])
                Permission.objects.create(codename='read_%s' % entitytype, name='%s - read' % entitytype , content_type=content_type[0])
                Permission.objects.create(codename='delete_%s' % entitytype, name='%s - delete' % entitytype , content_type=content_type[0])

    def load_resources(self, package_name, data_source=None):
        """
        Runs the setup.py file found in the package root

        """
        data_source = None if data_source == '' else data_source
        module = import_module('%s.setup' % package_name)
        load = getattr(module, 'load_resources')
        load(data_source) 

    def remove_resources(self, load_id):
        """
        Runs the resource_remover command found in package_utils

        """
        resource_remover.delete_resources(load_id)


    def start_livereload(self):
        from livereload import Server
        server = Server()
        for path in settings.STATICFILES_DIRS:
            server.watch(path)
        for path in settings.TEMPLATE_DIRS:
            server.watch(path)
        server.serve(port=settings.LIVERELOAD_PORT)
